# coding: utf-8

# isort: skip_file

from __future__ import annotations
from copy import deepcopy
from re import L
from typing import List, Union
import uuid
from supervisely.volume_annotation.volume_figure import VolumeFigure

from supervisely.project.project_meta import ProjectMeta
from supervisely._utils import take_with_default
from supervisely.io.fs import file_exists
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume_annotation.slice import Slice
from supervisely.volume_annotation.volume_tag_collection import VolumeTagCollection
from supervisely.volume_annotation.volume_object_collection import VolumeObjectCollection
from supervisely.volume_annotation.volume_object import VolumeObject
from supervisely.volume_annotation.volume_tag import VolumeTag
from supervisely.geometry.mask_3d import Mask3D
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.volume_annotation.plane import Plane
from supervisely.volume_annotation.constants import (
    NAME,
    TAGS,
    OBJECTS,
    KEY,
    VOLUME_ID,
    VOLUME_META,
    PLANES,
    SPATIAL_FIGURES,
)
from supervisely.io.fs import get_file_name
from supervisely.io.json import dump_json_file


class VolumeAnnotation:
    """Annotation for a 3D volume: objects, tags, planes, spatial figures. Immutable."""

    def __init__(
        self,
        volume_meta,
        objects=None,
        plane_sagittal=None,
        plane_coronal=None,
        plane_axial=None,
        tags=None,
        spatial_figures=None,
        key=None,
    ):
        """
        Annotation for a 3D volume.

        :param volume_meta: Volume metadata (dimensions, spacing, origin, etc.).
        :type volume_meta: dict
        :param objects: Collection of VolumeObjects.
        :type objects: :class:`~supervisely.volume_annotation.volume_object_collection.VolumeObjectCollection`, optional
        :param plane_sagittal: Sagittal plane. Default Plane created if not provided.
        :type plane_sagittal: :class:`~supervisely.volume_annotation.plane.Plane`, optional
        :param plane_coronal: Coronal plane. Default Plane created if not provided.
        :type plane_coronal: :class:`~supervisely.volume_annotation.plane.Plane`, optional
        :param plane_axial: Axial plane. Default Plane created if not provided.
        :type plane_axial: :class:`~supervisely.volume_annotation.plane.Plane`, optional
        :param tags: Collection of VolumeTags.
        :type tags: :class:`~supervisely.volume_annotation.volume_tag_collection.VolumeTagCollection`, optional
        :param spatial_figures: List of VolumeFigures (e.g. Mask3D).
        :type spatial_figures: List[:class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`], optional
        :param key: UUID key. Auto-generated if not provided.
        :type key: uuid.UUID, optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                volume_ann = sly.VolumeAnnotation(volume_meta)

                obj_class = sly.ObjClass('brain', sly.Rectangle)
                volume_obj = sly.VolumeObject(obj_class)
                vol_tag = sly.VolumeTag(sly.TagMeta('brain_tag', sly.TagValueType.ANY_STRING), value='human')
                volume_ann = sly.VolumeAnnotation(
                    volume_meta,
                    sly.VolumeObjectCollection([volume_obj]),
                    tags=sly.VolumeTagCollection([vol_tag]),
                )
        """
        self._volume_meta = volume_meta
        self._tags = take_with_default(tags, VolumeTagCollection())
        self._objects = take_with_default(objects, VolumeObjectCollection())
        self._key = take_with_default(key, uuid.uuid4())

        self._plane_sagittal = take_with_default(
            plane_sagittal,
            Plane(Plane.SAGITTAL, volume_meta=volume_meta),
        )
        self._plane_coronal = take_with_default(
            plane_coronal,
            Plane(Plane.CORONAL, volume_meta=volume_meta),
        )
        self._plane_axial = take_with_default(
            plane_axial,
            Plane(Plane.AXIAL, volume_meta=volume_meta),
        )

        self._spatial_figures = take_with_default(spatial_figures, [])
        self.validate_figures_bounds()

    @property
    def volume_meta(self) -> dict:
        """
        Volume meta data.

        :returns: Sagittal plane of the volume.
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                vol_ann = sly.VolumeAnnotation(volume_meta)
                volume_meta = vol_ann.volume_meta
        """

        return deepcopy(self._volume_meta)

    @property
    def plane_sagittal(self) -> Plane:
        """
        Sagital plane of the volume.

        :returns: Sagittal plane of the volume.
        :rtype: :class:`~supervisely.volume_annotation.plane.Plane`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                vol_ann = sly.VolumeAnnotation(volume_meta)
                plane_sagittal = vol_ann.plane_sagittal
        """

        return self._plane_sagittal

    @property
    def plane_coronal(self) -> Plane:
        """
        Coronal plane of the volume.

        :returns: Coronal plane of the volume.
        :rtype: :class:`~supervisely.volume_annotation.plane.Plane`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                vol_ann = sly.VolumeAnnotation(volume_meta)
                plane_coronal = vol_ann.plane_coronal
        """

        return self._plane_coronal

    @property
    def plane_axial(self) -> Plane:
        """
        Axial plane of the volume.

        :returns: Axial plane of the volume.
        :rtype: :class:`~supervisely.volume_annotation.plane.Plane`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                vol_ann = sly.VolumeAnnotation(volume_meta)
                plane_axial = vol_ann.plane_axial
        """

        return self._plane_axial

    @property
    def objects(self) -> VolumeObjectCollection:
        """
        VolumeAnnotation objects.

        :returns: VolumeObjectCollection object
        :rtype: :class:`~supervisely.volume_annotation.volume_object_collection.VolumeObjectCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                path = "/Users/Downloads/volumes/Demo volumes_ds1_CTChest.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)

                # VolumeObjectCollection
                obj_class_heart = sly.ObjClass('heart', sly.Rectangle)
                volume_obj_heart = sly.VolumeObject(obj_class_heart)
                objects = sly.VolumeObjectCollection([volume_obj_heart])
                volume_ann = sly.VolumeAnnotation(volume_meta, objects)

                print(volume_ann.objects.to_json())
                # Output: [
                #     {
                #         "key": "2b5d70baa5a74d06a525b950b5f2b756",
                #         "classTitle": "heart",
                #         "tags": []
                #     }
                # ]
        """

        return self._objects

    @property
    def tags(self) -> VolumeTagCollection:
        """
        VolumeTag objects.

        :returns: VolumeTagCollection object
        :rtype: :class:`~supervisely.volume_annotation.volume_tag_collection.VolumeTagCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                vol_ann = sly.VolumeAnnotation(volume_meta)
                tags = vol_ann.tags
        """

        return self._tags

    @property
    def spatial_figures(self) -> List[VolumeFigure]:
        """
        Get a list of spatial figures.

        :returns: List of spatial figures from :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation` object.
        :rtype: List[:class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                vol_ann = sly.VolumeAnnotation(volume_meta)
                spatial_figures = vol_ann.spatial_figures
        """

        return self._spatial_figures

    @property
    def figures(self) -> List[VolumeFigure]:
        """
        VolumeFigure objects.

        :returns: List of VolumeFigure objects from VolumeAnnotation object.
        :rtype: List[:class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                vol_ann = sly.VolumeAnnotation(volume_meta)
                figures = vol_ann.figures
        """

        all_figures = []
        for plane in [self.plane_sagittal, self.plane_coronal, self.plane_axial]:
            all_figures.extend(plane.figures)
        return all_figures

    def key(self) -> str:
        """
        Volume annotation key value.

        :returns: Key value of VolumeAnnotation object.
        :rtype: str

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                vol_ann = sly.VolumeAnnotation(volume_meta)
                key = vol_ann.key()
        """
        return self._key

    def validate_figures_bounds(self):
        """
        Checks if all slices in each plane contains figures.

        :raises :class:`~supervisely.video_annotation.video_figure.OutOfImageBoundsException`: if figure is out of slices images bounds
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                plane_axial = sly.Plane(sly.Plane.AXIAL, frames, volume_meta=volume_meta)
                volume_ann = sly.VolumeAnnotation(volume_meta, objects, plane_axial=plane_axial)
                volume_ann.validate_figures_bounds()
        """

        self.plane_sagittal.validate_figures_bounds()
        self.plane_coronal.validate_figures_bounds()
        self.plane_axial.validate_figures_bounds()

    def is_empty(self) -> bool:
        """
        Check whether volume annotation contains objects or tags, or not.

        :returns: True if volume annotation is empty, False otherwise.
        :rtype: bool

        :Usage exmaple:

         .. code-block:: python

            import supervisely as sly
            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            vol_ann = sly.VolumeAnnotation(volume_meta)

            is_empty = vol_ann.is_empty()
        """

        if len(self.objects) == 0 and len(self.tags) == 0:
            return True
        else:
            return False

    def clone(
        self,
        volume_meta=None,
        objects=None,
        plane_sagittal=None,
        plane_coronal=None,
        plane_axial=None,
        tags=None,
        spatial_figures=None,
    ):
        """
        Makes a copy of VolumeAnnotation with new fields, if fields are given, otherwise it will use fields of the original VolumeAnnotation.

        :param volume_meta: Metadata of the volume.
        :type volume_meta: dict
        :param objects: VolumeObjectCollection object.
        :type objects: :class:`~supervisely.volume_annotation.volume_object_collection.VolumeObjectCollection`, optional
        :param plane_sagittal: Sagittal plane of the volume.
        :type plane_sagittal: Plane object, optional
        :param plane_coronal: Coronal plane of the volume.
        :type plane_coronal: Plane object, optional
        :param plane_axial: Axial plane of the volume.
        :type plane_axial: Plane object, optional
        :param tags: VolumeTagCollection object.
        :type tags: :class:`~supervisely.volume_annotation.volume_tag_collection.VolumeTagCollection`, optional
        :param spatial_figures: List of spatial figures associated with the volume.
        :type spatial_figures: List[:class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`], optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                volume_ann = sly.VolumeAnnotation(volume_meta)

                obj_class_heart = sly.ObjClass('heart', sly.Rectangle)
                video_obj_heart = sly.VolumeObject(obj_class_heart)
                new_objects = sly.VolumeObjectCollection([volume_obj_heart])
                new_volume_ann = volume_ann.clone(objects=new_objects)
        """

        return VolumeAnnotation(
            volume_meta=take_with_default(volume_meta, self.volume_meta),
            objects=take_with_default(objects, self.objects),
            plane_sagittal=take_with_default(plane_sagittal, self.plane_sagittal),
            plane_coronal=take_with_default(plane_coronal, self.plane_coronal),
            plane_axial=take_with_default(plane_axial, self.plane_axial),
            tags=take_with_default(tags, self.tags),
            spatial_figures=take_with_default(spatial_figures, self.spatial_figures),
        )

    @classmethod
    def from_json(
        cls,
        data: dict,
        project_meta: ProjectMeta,
        key_id_map: KeyIdMap = None,
        spatial_geometry_paths: Union[list, dict] = None,
    ):
        """
        Convert a json dict to VolumeAnnotation.

        :param data: Volume annotation in json format as a dict.
        :type data: dict
        :param project_meta: Input ProjectMeta object.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :param spatial_geometry_paths: Optional. Can be either:
            - a list of file paths to spatial geometry files, where each file name should match either the figure's id or the hex value of its key,
            - or a dict mapping figure ids (or keys) to their corresponding geometry file paths.
            Used to load 3D geometry for spatial figures.
        :type spatial_geometry_paths: list or dict, optional
        :returns: VolumeAnnotation object
        :rtype: :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                from supervisely.video_annotation.key_id_map import KeyIdMap

                meta = sly.ProjectMeta()
                key_id_map = KeyIdMap()

                ann_json = {
                    "key": "56107223943346e5900fc256b8dcd7f0",
                    "objects": [],
                    "planes": [
                        { "name": "sagittal", "normal": { "x": 1, "y": 0, "z": 0 }, "slices": [] },
                        { "name": "coronal", "normal": { "x": 0, "y": 1, "z": 0 }, "slices": [] },
                        { "name": "axial", "normal": { "x": 0, "y": 0, "z": 1 }, "slices": [] }
                    ],
                    "spatialFigures": [],
                    "tags": [],
                    "volumeMeta": {
                        "ACS": "RAS",
                        "channelsCount": 1,
                        "dimensionsIJK": { "x": 512, "y": 512, "z": 139 },
                        "directions": [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
                        "intensity": { "max": 3071.0, "min": -3024.0 },
                        "origin": [-194.238403081894, -217.5384061336518, -347.7500000000001],
                        "rescaleIntercept": 0,
                        "rescaleSlope": 1,
                        "spacing": [0.7617189884185793, 0.7617189884185793, 2.5],
                        "windowCenter": 23.5,
                        "windowWidth": 6095.0
                    }
                }

                ann = sly.VolumeAnnotation.from_json(ann_json, meta, key_id_map)
        """

        volume_key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()
        if key_id_map is not None:
            key_id_map.add_video(volume_key, data.get(VOLUME_ID, None))

        volume_meta = data[VOLUME_META]

        tags = VolumeTagCollection.from_json(data[TAGS], project_meta.tag_metas, key_id_map)
        objects = VolumeObjectCollection.from_json(data[OBJECTS], project_meta, key_id_map)

        plane_sagittal = None
        plane_coronal = None
        plane_axial = None
        for plane_json in data[PLANES]:
            if plane_json[NAME] == Plane.SAGITTAL:
                plane_sagittal = Plane.from_json(
                    plane_json,
                    Plane.SAGITTAL,
                    objects,
                    volume_meta=volume_meta,
                    key_id_map=key_id_map,
                )
            elif plane_json[NAME] == Plane.CORONAL:
                plane_coronal = Plane.from_json(
                    plane_json,
                    Plane.CORONAL,
                    objects,
                    volume_meta=volume_meta,
                    key_id_map=key_id_map,
                )
            elif plane_json[NAME] == Plane.AXIAL:
                plane_axial = Plane.from_json(
                    plane_json,
                    Plane.AXIAL,
                    objects,
                    volume_meta=volume_meta,
                    key_id_map=key_id_map,
                )
            else:
                raise RuntimeError(f"Unknown plane name {plane_json[NAME]}")

        spatial_figures = []
        for figure_json in data.get(SPATIAL_FIGURES, []):
            figure = VolumeFigure.from_json(
                figure_json,
                objects,
                plane_name=None,
                slice_index=None,
                key_id_map=key_id_map,
            )
            if spatial_geometry_paths:
                figure_id = figure_json["id"]
                if isinstance(spatial_geometry_paths, list):
                    spatial_geometry_paths = {
                        get_file_name(path): path for path in spatial_geometry_paths
                    }
                    geometry_path = spatial_geometry_paths.get(
                        figure.key().hex
                    ) or spatial_geometry_paths.get(str(figure_id))
                elif isinstance(spatial_geometry_paths, dict):
                    geometry_path = spatial_geometry_paths.get(figure_id, None)
                else:
                    raise ValueError(
                        f"spatial_geometry_paths should be either a list or a dict. Got: {type(spatial_geometry_paths)}"
                    )

                if geometry_path is not None:
                    if not file_exists(geometry_path):
                        raise RuntimeError(
                            f"Geometry file {geometry_path} for figure {figure_id} does not exist."
                        )
                    mask3d = Mask3D.create_from_file(geometry_path)
                    mask3d.sly_id = figure_id
                    figure._set_3d_geometry(mask3d)

            spatial_figures.append(figure)

        return cls(
            volume_meta=volume_meta,
            objects=objects,
            plane_sagittal=plane_sagittal,
            plane_coronal=plane_coronal,
            plane_axial=plane_axial,
            tags=tags,
            spatial_figures=spatial_figures,
            key=volume_key,
        )

    def to_json(self, key_id_map: KeyIdMap = None) -> dict:
        """
        Convert the VolumeAnnotation to a json dict.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: Volume annotation in json format as a dict.
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                from supervisely.video_annotation.key_id_map import KeyIdMap

                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                volume_ann = sly.VolumeAnnotation(volume_meta)

                print(volume_ann.to_json())
                # Output: {
                # {
                #     "key": "56107223943346e5900fc256b8dcd7f0",
                #     "objects": [],
                #     "planes": [
                #         { "name": "sagittal", "normal": { "x": 1, "y": 0, "z": 0 }, "slices": [] },
                #         { "name": "coronal", "normal": { "x": 0, "y": 1, "z": 0 }, "slices": [] },
                #         { "name": "axial", "normal": { "x": 0, "y": 0, "z": 1 }, "slices": [] }
                #     ],
                #     "spatialFigures": [],
                #     "tags": [],
                #     "volumeMeta": {
                #         "ACS": "RAS",
                #         "channelsCount": 1,
                #         "dimensionsIJK": { "x": 512, "y": 512, "z": 139 },
                #         "directions": [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
                #         "intensity": { "max": 3071.0, "min": -3024.0 },
                #         "origin": [-194.238403081894, -217.5384061336518, -347.7500000000001],
                #         "rescaleIntercept": 0,
                #         "rescaleSlope": 1,
                #         "spacing": [0.7617189884185793, 0.7617189884185793, 2.5],
                #         "windowCenter": 23.5,
                #         "windowWidth": 6095.0
                #     }
                # }
        """

        res_json = {
            VOLUME_META: self.volume_meta,
            KEY: self.key().hex,
            TAGS: self.tags.to_json(key_id_map),
            OBJECTS: self.objects.to_json(key_id_map),
            PLANES: [
                self.plane_sagittal.to_json(),
                self.plane_coronal.to_json(),
                self.plane_axial.to_json(),
            ],
            SPATIAL_FIGURES: [figure.to_json(key_id_map) for figure in self.spatial_figures],
        }

        if key_id_map is not None:
            volume_id = key_id_map.get_video_id(self.key())
            if volume_id is not None:
                res_json[VOLUME_ID] = volume_id

        return res_json

    def dump_json(self, path: str, key_id_map: KeyIdMap = None) -> None:
        """
        Save the VolumeAnnotation to a json file.

        :param path: Path to the json file.
        :type path: str
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                path = "/home/admin/work/volumes/vol_01.nrrd"
                volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                volume_ann = sly.VolumeAnnotation(volume_meta)
                volume_ann.dump_json("/home/admin/work/volumes/vol_01.json")
        """
        simple_dict = self.to_json(key_id_map)
        dump_json_file(simple_dict, path)

    def add_objects(
        self, objects: Union[List[VolumeObject], VolumeObjectCollection]
    ) -> VolumeAnnotation:
        """
        Add new objects to a VolumeAnnotation object.

        :param objects: New volume objects.
        :type objects: List[VolumeObject] or VolumeObjectCollection
        :returns: A :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation` object containing the original and new volume objects.
        :rtype: :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                path = "/vol_01.nrrd"
                _, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                volume_ann = sly.VolumeAnnotation(volume_meta)
                obj_class_heart = sly.ObjClass('heart', sly.Mask3D)
                volume_obj_heart = sly.VolumeObject(obj_class_heart)
                volume_ann = volume_ann.add_objects([volume_obj_heart])

        """

        sf_figures = []
        for volume_object in objects:
            if volume_object.obj_class.geometry_type in (Mask3D, AnyGeometry):
                if isinstance(volume_object.figure.geometry, Mask3D):
                    sf_figures.append(volume_object.figure)

        collection = self.objects.add_items(objects)
        new_ann = self.clone(objects=collection)
        new_ann.spatial_figures.extend(sf_figures)
        return new_ann

    def remove_objects(self, keys: Union[List[uuid.UUID], uuid.UUID]) -> VolumeAnnotation:
        """
        Remove annotation objects from a VolumeAnnotation object.

        :param keys: List of object keys or single object key.
        :type keys: List[uuid.UUID] or uuid.UUID
        :returns: A VolumeAnnotation object containing the original volume objects without the removed objects.
        :rtype: :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                path = "/vol_01.nrrd"
                _, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                volume_ann = sly.VolumeAnnotation(volume_meta)
                obj_class_heart = sly.ObjClass('heart', sly.Mask3D)
                volume_obj_heart = sly.VolumeObject(obj_class_heart)
                obj_class_tumor = sly.ObjClass('tumor', sly.Mask3D)
                volume_obj_tumor = sly.VolumeObject(obj_class_tumor)
                volume_ann = volume_ann.add_objects([volume_obj_heart, volume_obj_tumor])
                volume_ann = volume_ann.remove_objects(volume_obj_heart.key())

        """
        if not isinstance(keys, list):
            keys = [keys]

        spatial_figures_to_del = []
        for ann_object in self.objects:
            if ann_object.key() in keys and ann_object.obj_class.geometry_type in (
                Mask3D,
                AnyGeometry,
            ):
                for figure in self.spatial_figures:
                    if figure.parent_object == ann_object:
                        spatial_figures_to_del.append(figure)

        for spatial_figure in spatial_figures_to_del:
            self.spatial_figures.remove(spatial_figure)

        collection = self.objects.remove_items(keys)
        new_ann = self.clone(objects=collection)
        return new_ann

    def add_tags(self, tags: Union[List[VolumeTag], VolumeTagCollection]) -> VolumeAnnotation:
        """
        Add new VolumeTags to a VolumeAnnotation object.

        :param tags: New VolumeTags.
        :type tags: List[:class:`~supervisely.volume_annotation.volume_tag.VolumeTag`] or :class:`~supervisely.volume_annotation.volume_tag_collection.VolumeTagCollection`
        :returns: A :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation` object containing the original and new volume tags.
        :rtype: :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                path = "/vol_01.nrrd"
                _, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                volume_ann = sly.VolumeAnnotation(volume_meta)
                brain_meta = sly.TagMeta('brain_tag', sly.TagValueType.ANY_STRING)
                vol_tag = sly.VolumeTag(brain_meta, value='human')
                volume_ann = volume_ann.add_tags([vol_tag])

        """
        collection = self.tags.add_items(tags)
        return self.clone(tags=collection)

    def remove_tags(self, keys: Union[List[uuid.UUID], uuid.UUID]) -> VolumeAnnotation:
        """
        Remove VolumeTags from a VolumeAnnotation object.

        :param keys: List of VolumeTag keys or single tag key.
        :type keys: List[uuid.UUID] or uuid.UUID
        :returns: A :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation` object containing the original VolumeTags without the removed VolumeTags.
        :rtype: :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                path = "/vol_01.nrrd"
                _, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
                volume_ann = sly.VolumeAnnotation(volume_meta)
                brain_meta = sly.TagMeta('brain_tag', sly.TagValueType.ANY_STRING)
                vol_tag = sly.VolumeTag(brain_meta, value='human')
                volume_ann = volume_ann.add_tags([vol_tag])
                volume_ann = volume_ann.remove_tags(vol_tag.key())

        """
        if not isinstance(keys, list):
            keys = [keys]
        collection = self.tags.remove_items(keys)
        return self.clone(tags=collection)



# coding: utf-8

# docs
import uuid
from typing import Optional

from supervisely.video_annotation.video_object import VideoObject
from supervisely.annotation.label import LabelJsonFields
from supervisely.project.project_meta import ProjectMeta
from supervisely.video_annotation.constants import KEY, ID
from supervisely.pointcloud_annotation.pointcloud_tag_collection import PointcloudTagCollection
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.geometry.constants import LABELER_LOGIN, UPDATED_AT, CREATED_AT, CLASS_ID


class PointcloudObject(VideoObject):
    """Object in point cloud annotation (obj_class + tags). Uses VideoObject constructor. Immutable."""

    tag_collection_type = PointcloudTagCollection

    @classmethod
    def from_json(cls, data, project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap] = None):
        """
        Convert PointcloudObject from json format to PointcloudObject class object. Raise error if object class name is not found in the given project meta.

        :param data: PointcloudObject in json format.
        :type data: dict
        :param project_meta: Project metadata.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param key_id_map: Key ID map.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: Pointcloud object.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_object.PointcloudObject`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

                key_id_map = KeyIdMap()
                project_id = 19441
                obj_class_car = sly.ObjClass('car', Cuboid3d)
                project_meta = sly.ProjectMeta([obj_class_car])
                api.project.update_meta(id=project_id, meta=project_meta)
                pointcloud_obj_car = sly.PointcloudObject(obj_class_car)
                pointcloud_obj_car_json = pointcloud_obj_car.to_json()
                new_pointcloud_obj_car = sly.PointcloudObject.from_json(
                    data=pointcloud_obj_car_json,
                    project_meta=project_meta,
                    key_id_map=key_id_map
                )
                print(new_pointcloud_obj_car)
                # <supervisely.pointcloud_annotation.pointcloud_object.PointcloudObject object at 0x7f97e0ba8ed0>
        """

        obj_class_name = data[LabelJsonFields.OBJ_CLASS_NAME]
        obj_class = project_meta.get_obj_class(obj_class_name)
        if obj_class is None:
            raise RuntimeError(
                f"Failed to deserialize a object from JSON: class name {obj_class_name!r} "
                f"was not found in the given project meta."
            )

        object_id = data.get(ID, None)

        existing_key = None
        if object_id is not None and key_id_map is not None:
            existing_key = key_id_map.get_object_key(object_id)
        json_key = uuid.UUID(data[KEY]) if KEY in data else None
        if (existing_key is not None) and (json_key is not None) and (existing_key != json_key):
            raise RuntimeError(
                "Object id = {!r}: existing_key {!r} != json_key {!r}".format(
                    object_id, existing_key, json_key
                )
            )

        if existing_key is not None:
            key = existing_key
        elif json_key is not None:
            key = json_key
        else:
            key = uuid.uuid4()

        if key_id_map is not None and existing_key is None:
            key_id_map.add_object(key, object_id)

        class_id = data.get(CLASS_ID, None)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)

        return cls(
            obj_class=obj_class,
            key=key,
            tags=cls.tag_collection_type.from_json(
                data[LabelJsonFields.TAGS], project_meta.tag_metas
            ),
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )



import os

import cv2


def build_segmentation_loader(
    pred_dir,
    gt_dir,
    num_workers=0,
    pred_label_map=None,
    gt_label_map=None,
):
    import torch  # pylint: disable=import-error

    class SegmentationLoader(torch.utils.data.Dataset):
        """PyTorch Dataset yielding (prediction, ground-truth) mask pairs from pred_dir and gt_dir."""

        def __init__(
            self,
            pred_dir,
            gt_dir,
            pred_label_map=None,
            gt_label_map=None,
        ):
            self.pred_dir = pred_dir
            self.pred_files = sorted(os.listdir(self.pred_dir))
            self.gt_dir = gt_dir
            self.gt_files = sorted(os.listdir(self.gt_dir))
            if len(self.pred_files) != len(self.gt_files):
                raise RuntimeError(
                    f"Number of predictions({len(self.pred_files)}) and ground-truth annotations ({len(self.gt_files)}) do not match!"
                )
            self.pred_label_map = pred_label_map
            self.gt_label_map = gt_label_map

        def __len__(self):
            return len(self.pred_files)

        def __getitem__(self, index):
            pred_path = os.path.join(self.pred_dir, self.pred_files[index])
            gt_path = os.path.join(self.gt_dir, self.gt_files[index])

            if self.pred_label_map:
                pred = self.pred_label_map(pred)
            else:
                pred = cv2.imread(pred_path, cv2.IMREAD_UNCHANGED)

            if self.gt_label_map:
                gt = self.gt_label_map(gt)
            else:
                gt = cv2.imread(gt_path, cv2.IMREAD_UNCHANGED)

            img_name = self.gt_files[index]
            return (pred, gt, img_name)

    loader = SegmentationLoader(
        pred_dir=pred_dir,
        gt_dir=gt_dir,
        pred_label_map=pred_label_map,
        gt_label_map=gt_label_map,
    )
    return torch.utils.data.DataLoader(
        loader,
        batch_size=1,
        num_workers=num_workers,
        collate_fn=lambda l: (l[0][0], l[0][1], l[0][2]),
    )



import numpy as np
from typing import Dict, Tuple

class DetectionMetrics:
    """Detection metrics for Evaluator (IoU, precision, recall, F1, assistance score)."""

    def __init__(self, iou_min: float = 0.7, iou_perfect: float = 0.9):
        """
        :param iou_min: Min IoU for match.
        :type iou_min: float
        :param iou_perfect: IoU for perfect match.
        :type iou_perfect: float
        """
        self.iou_min = iou_min
        self.iou_perfect = iou_perfect
        self.score_weights = {
            'tp': 1.0,
            'fs': 0.5,
            'fc': 0.7,
            'fp': -0.1,
            'fn': 0.0
        }

    def iou_matrix(self, pred_boxes: np.ndarray, gt_boxes: np.ndarray) -> np.ndarray:
        """Compute IoU matrix between predicted and GT boxes (XYXY format)."""
        if len(pred_boxes) == 0 or len(gt_boxes) == 0:
            return np.zeros((len(pred_boxes), len(gt_boxes)))
        
        pred_boxes = np.array(pred_boxes, dtype=np.float32)
        gt_boxes = np.array(gt_boxes, dtype=np.float32)
        
        # Expand for broadcasting: [N, 1, 4] and [1, M, 4]
        pred_expanded = pred_boxes[:, np.newaxis, :]
        gt_expanded = gt_boxes[np.newaxis, :, :]
        
        # Intersection coordinates
        x1_inter = np.maximum(pred_expanded[:, :, 0], gt_expanded[:, :, 0])
        y1_inter = np.maximum(pred_expanded[:, :, 1], gt_expanded[:, :, 1])
        x2_inter = np.minimum(pred_expanded[:, :, 2], gt_expanded[:, :, 2])
        y2_inter = np.minimum(pred_expanded[:, :, 3], gt_expanded[:, :, 3])
        
        # Intersection area
        inter_w = np.maximum(0.0, x2_inter - x1_inter)
        inter_h = np.maximum(0.0, y2_inter - y1_inter)
        intersection = inter_w * inter_h
        
        # Box areas
        pred_area = (pred_boxes[:, 2] - pred_boxes[:, 0]) * (pred_boxes[:, 3] - pred_boxes[:, 1])
        gt_area = (gt_boxes[:, 2] - gt_boxes[:, 0]) * (gt_boxes[:, 3] - gt_boxes[:, 1])
        
        # Union = area1 + area2 - intersection
        union = pred_area[:, np.newaxis] + gt_area[np.newaxis, :] - intersection
        
        # IoU with division by zero protection
        iou = intersection / np.maximum(union, 1e-10)
        
        return iou

    def match_boxes(self, pred_boxes, pred_labels, gt_boxes, gt_labels, iou_threshold: float):
        """Match predictions to GT boxes using Hungarian algorithm."""
        from scipy.optimize import linear_sum_assignment

        pred_matches = {}
        gt_matches = {}
        if len(pred_boxes) == 0 or len(gt_boxes) == 0:
            return pred_matches, gt_matches
        
        iou_mat = self.iou_matrix(pred_boxes, gt_boxes)
        valid = iou_mat >= iou_threshold
        cost_mat = -iou_mat.copy()
        cost_mat[~valid] = 0
        
        pred_idx, gt_idx = linear_sum_assignment(cost_mat)
        for p, g in zip(pred_idx, gt_idx):
            iou = iou_mat[p, g]
            if iou >= iou_threshold:
                class_match = pred_labels[p] == gt_labels[g]
                pred_matches[p] = (g, iou, class_match)
                gt_matches[g] = (p, iou, class_match)
        
        return pred_matches, gt_matches

    def assistance_score(self, pred_boxes, pred_labels, gt_boxes, gt_labels) -> Dict:
        """Compute annotation assistance score and components."""
        num_gt = len(gt_boxes)
        if num_gt == 0:
            return {'assistance_score': 0.0, 'tp':0, 'fs':0, 'fc':0, 'fp':0, 'fn':0, 'num_gt':0}

        pred_matches, gt_matches = self.match_boxes(pred_boxes, pred_labels, gt_boxes, gt_labels, self.iou_min)
        
        tp = fs = fc = fp = 0
        for i in range(len(pred_boxes)):
            if i in pred_matches:
                g_idx, iou, class_match = pred_matches[i]
                if iou >= self.iou_perfect and class_match: tp += 1
                elif self.iou_min <= iou < self.iou_perfect and class_match: fs += 1
                elif iou >= self.iou_perfect and not class_match: fc += 1
                else: fp += 1
            else:
                fp += 1
        fn = num_gt - len(gt_matches)
        raw_score = (
            tp*self.score_weights['tp'] +
            fs*self.score_weights['fs'] +
            fc*self.score_weights['fc'] +
            fp*self.score_weights['fp'] +
            fn*self.score_weights['fn']
        )

        score = max(0.0, raw_score / num_gt)

        return {'assistance_score': score, 'tp': tp, 'fs': fs, 'fc': fc, 'fp': fp, 'fn': fn, 'num_gt': num_gt}

    def detection_metrics(self, pred_boxes, pred_labels, gt_boxes, gt_labels, iou_threshold: float) -> Dict:
        """Compute precision, recall, F1 at given IoU threshold."""
        pred_matches, gt_matches = self.match_boxes(pred_boxes, pred_labels, gt_boxes, gt_labels, iou_threshold)
        tp = sum(1 for _, (_, _, c) in pred_matches.items() if c)
        precision = tp / len(pred_boxes) if len(pred_boxes)>0 else 0.0
        recall = tp / len(gt_boxes) if len(gt_boxes)>0 else 0.0
        f1 = 2*precision*recall/(precision+recall) if (precision+recall)>0 else 0.0
        return {'precision': precision, 'recall': recall, 'f1_score': f1}

    def all_metrics(self, pred_boxes, pred_labels, gt_boxes, gt_labels) -> Dict:
        """Compute all metrics together (assistance + detection)."""
        metrics = self.assistance_score(pred_boxes, pred_labels, gt_boxes, gt_labels)
        metrics_min = self.detection_metrics(pred_boxes, pred_labels, gt_boxes, gt_labels, self.iou_min)
        metrics_perf = self.detection_metrics(pred_boxes, pred_labels, gt_boxes, gt_labels, self.iou_perfect)

        metrics.update({
            f'precision_iou_{self.iou_min}': metrics_min['precision'],
            f'recall_iou_{self.iou_min}': metrics_min['recall'],
            f'f1_score_iou_{self.iou_min}': metrics_min['f1_score'],
            f'precision_iou_{self.iou_perfect}': metrics_perf['precision'],
            f'recall_iou_{self.iou_perfect}': metrics_perf['recall'],
            f'f1_score_iou_{self.iou_perfect}': metrics_perf['f1_score']
        })
        return metrics



from __future__ import annotations

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    CollapseWidget,
    MarkdownWidget,
    NotificationWidget,
)


class ReliabilityDiagram(DetectionVisMetric):
    """Calibration curve and ECE/MCE for detection reports."""

    MARKDOWN_CALIBRATION_SCORE = "calibration_score"
    MARKDOWN_CALIBRATION_SCORE_2 = "calibration_score_2"
    MARKDOWN_RELIABILITY_DIAGRAM = "reliability_diagram"
    NOTIFICATION = "reliability_diagram"
    CHART = "reliability_diagram"

    @property
    def md_calibration_score(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_calibration_score_1
        text = text.format(self.vis_texts.definitions.confidence_score)
        return MarkdownWidget(self.MARKDOWN_CALIBRATION_SCORE, "Calibration Score", text)

    @property
    def collapse_tip(self) -> CollapseWidget:
        md = MarkdownWidget(
            "what_is_calibration",
            "What is calibration?",
            self.vis_texts.markdown_what_is_calibration,
        )
        return CollapseWidget([md])

    @property
    def md_calibration_score_2(self) -> MarkdownWidget:
        return MarkdownWidget(
            self.MARKDOWN_CALIBRATION_SCORE_2,
            "",
            self.vis_texts.markdown_calibration_score_2,
        )

    @property
    def md_reliability_diagram(self) -> MarkdownWidget:
        return MarkdownWidget(
            self.MARKDOWN_RELIABILITY_DIAGRAM,
            "Reliability Diagram",
            self.vis_texts.markdown_reliability_diagram,
        )

    @property
    def notification(self) -> NotificationWidget:
        title, _ = self.vis_texts.notification_ece.values()
        return NotificationWidget(
            self.NOTIFICATION,
            title.format(self.eval_result.mp.m_full.expected_calibration_error().round(4)),
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget(self.CHART, self._get_figure())

    @property
    def collapse(self) -> CollapseWidget:
        md = MarkdownWidget(
            "markdown_calibration_curve_interpretation",
            "How to interpret the Calibration curve",
            self.vis_texts.markdown_calibration_curve_interpretation,
        )
        return CollapseWidget([md])

    def _get_figure(self):  # -> go.Figure:
        import plotly.graph_objects as go  # pylint: disable=import-error

        # Calibration curve (only positive predictions)
        true_probs, pred_probs = self.eval_result.mp.m_full.calibration_curve()

        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=pred_probs,
                y=true_probs,
                mode="lines+markers",
                name="Calibration plot (Model)",
                line=dict(color="blue"),
                marker=dict(color="blue"),
            )
        )
        fig.add_trace(
            go.Scatter(
                x=[0, 1],
                y=[0, 1],
                mode="lines",
                name="Perfectly calibrated",
                line=dict(color="orange", dash="dash"),
            )
        )

        fig.update_layout(
            xaxis_title="Confidence Score",
            yaxis_title="Fraction of True Positives",
            legend=dict(x=0.6, y=0.1),
            xaxis=dict(range=[0, 1]),
            yaxis=dict(range=[0, 1]),
            width=700,
            height=500,
        )
        fig.update_traces(
            hovertemplate="Confidence Score: %{x:.2f}<br>Fraction of True Positives: %{y:.2f}<extra></extra>"
        )
        return fig



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from copy import deepcopy
from typing import List

from supervisely import logger


class ImageSlider(Widget):
    """Horizontal slider of image previews with optional extra examples per item; emits selection changes."""

    def __init__(
        self,
        previews: List[str] = None,
        examples: List[List[str]] = None,
        combined_data: List[dict] = None,
        height: int = 200,
        selectable: bool = True,
        preview_idx: int = None,
        preview_url: str = None,
        widget_id: str = None,
    ):
        """
        :param previews: List of image URLs.
        :type previews: List[str]
        :param examples: List of lists of image URLs.
        :type examples: List[List[str]]
        :param combined_data: List of dicts with 'moreExamples' and 'preview' keys.
        :type combined_data: List[dict]
        :param height: Height in pixels.
        :type height: int
        :param selectable: If True, the slider is selectable.
        :type selectable: bool
        :param preview_idx: Index of the preview to select.
        :type preview_idx: int
        :param preview_url: URL of the preview to select.
        :type preview_url: str
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional

        Input parameters format examples:

            previews = ["https://i.imgur.com/1Ys222.png", "https://i.imgur.com/2Yj2QjQ.png",]

            examples = [["https://i.imgur.com/1Ys222.png", "https://i.imgur.com/3Yd243.png"], ...]

            combined_data = [
                {
                    "moreExamples": [
                        "https://i.imgur.com/1Ys222.png",
                        "https://i.imgur.com/3Yd243.png",
                    ],
                    "preview": "https://i.imgur.com/2Yj2QjQ.png"
                },
                ...
            ]
        """

        self._data, self._image_url_to_idx = self._process_data(previews, examples, combined_data)
        self._height = height
        self._selectable = selectable
        self._selected_idx = None
        self._selected = None

        if preview_idx is not None and preview_url is not None:
            raise ValueError("You can't specify both 'preview_idx' and 'preview_url'.")
        if preview_idx is not None:
            self._update_selected_by_idx(preview_idx)
        elif preview_url is not None:
            self._update_selected_by_url(preview_url)

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _process_data(
        self,
        previews: List[str] = None,
        examples: List[List[str]] = None,
        combined_data: List[dict] = None,
        start_idx: int = 0,
    ):
        data = []
        image_url_to_idx = {}
        if previews is not None and combined_data is not None:
            raise ValueError("You can't specify both 'previews' and 'combined_data'.")
        if previews is None and combined_data is None:
            logger.info("'previews' and 'combined_data' are not specified.")
            examples = []
            combined_data = []
        if previews is None and examples is not None:
            raise ValueError("You must specify 'previews' if you specify 'examples'.")
        if examples is not None and combined_data is not None:
            raise ValueError("You can't specify both 'examples' and 'combined_data'.")
        if examples is not None:
            if all([type(example) is not list for example in examples]):
                raise ValueError("Input examples must be a list of lists.")
            if len(examples) != len(previews):
                raise ValueError(
                    f"Length of 'examples' ({len(examples)}) must be equal to length of 'previews' ({len(previews)})."
                )
        if previews is not None:
            data = []
            if examples is None:
                examples = [[preview] for preview in previews]
            for idx, (preview, more_example) in enumerate(zip(previews, examples), start_idx):
                data.append(
                    {
                        "moreExamples": deepcopy(more_example),
                        "preview": preview,
                        "idx": idx,
                    }
                )
                image_url_to_idx[preview] = idx
        else:
            data = deepcopy(combined_data)
            for idx, item in enumerate(data, start_idx):
                image_url_to_idx[item["preview"]] = idx

        return data, image_url_to_idx

    def _update_selected_by_idx(self, idx: int):
        if type(idx) is not int:
            raise ValueError(f"Index must be an integer, not {type(idx)}")
        if 0 > idx >= len(self._data):
            raise ValueError(
                f'"Index {self._selected_idx} can`t be be greater than the length of input URLs list".'
            )
        self._selected_idx = idx
        self._selected = self._data[self._selected_idx]

        StateJson()[self.widget_id]["selected"] = self._selected
        StateJson().send_changes()

    def _update_selected_by_url(self, url: str):
        set_index = self._image_url_to_idx.get(url)
        if set_index is None:
            raise ValueError(f"There is no {url} url in input urls list")
        self._selected_idx = set_index
        self._selected = self._data[self._selected_idx]

        StateJson()[self.widget_id]["selected"] = self._selected
        StateJson().send_changes()

    def get_json_data(self):
        return {
            "data": self._data,
            "options": {
                "selectable": self._selectable,
                "height": f"{self._height}px",
            },
        }

    def get_json_state(self):
        return {"selected": self._selected}

    def get_selected_preview(self):
        selected = StateJson()[self.widget_id].get("selected")
        if selected is None:
            return None
        return selected["preview"]

    def set_selected_preview(self, value: str):
        self._update_selected_by_url(value)

    def get_selected_idx(self):
        selected = StateJson()[self.widget_id].get("selected")
        if selected is None:
            return None
        self._selected_idx = selected["idx"]
        return self._selected_idx

    def set_selected_idx(self, value: int):
        self._update_selected_by_idx(value)

    def get_selected_examples(self):
        selected = StateJson()[self.widget_id].get("selected")
        if selected is None:
            return None
        return selected["moreExamples"]

    @property
    def is_selectable(self):
        self._selectable = DataJson()[self.widget_id]["options"]["selectable"]
        return self._selectable

    def enable_selection(self):
        self._selectable = True
        DataJson()[self.widget_id]["options"]["selectable"] = self._selectable
        DataJson().send_changes()

    def disable_selection(self):
        self._selectable = False
        DataJson()[self.widget_id]["options"]["selectable"] = self._selectable
        DataJson().send_changes()

    def get_data_length(self):
        return len(self._data)

    def get_data(self):
        return self._data

    def set_data(
        self,
        previews: List[str] = None,
        examples: List[List[str]] = None,
        combined_data: List[dict] = None,
    ):
        self._data, self._image_url_to_idx = self._process_data(previews, examples, combined_data)
        DataJson()[self.widget_id]["data"] = self._data
        DataJson().send_changes()
        self._selected_idx = None
        self._selected = None
        StateJson()[self.widget_id]["selected"] = self._selected
        StateJson().send_changes()

    def append_data(
        self,
        previews: List[str] = None,
        examples: List[List[str]] = None,
        combined_data: List[dict] = None,
    ):
        data, url_to_idx = self._process_data(previews, examples, combined_data, len(self._data))
        self._data.extend(data)
        self._image_url_to_idx.update(url_to_idx)
        DataJson()[self.widget_id]["data"] = self._data
        DataJson().send_changes()



import os
import shutil
from pathlib import Path
from typing import Optional, Dict
from datetime import datetime
import supervisely as sly
from supervisely import logger
from supervisely.template.live_training.live_training_generator import LiveTrainingGenerator
import supervisely.io.json as sly_json
from supervisely.nn.live_training.helpers import ClassMap
import yaml


def upload_artifacts(
    api: sly.Api,
    session_info: dict,
    artifacts: dict,
) -> str:
    """
    Upload artifacts to Team Files and generate experiment report.

    :param api: Supervisely API object
    :type api: :class:`~supervisely.api.api.Api`
    :param session_info: Training session context
    :type session_info: dict
    :param artifacts: Framework-specific artifacts
    :type artifacts: dict
    :returns: URL to experiment report
    :rtype: str

    :Example:

        session_info: Training session context
            - team_id: Team ID
            - task_id: Task ID
            - project_id: Project ID
            - framework_name: Framework name
            - task_type: Task type
            - class_map: Model class map
            - start_time: Training start time string
            - train_size: Final dataset size
            - initial_samples: Number of initial samples


        artifacts: Framework-specific artifacts
            - checkpoint_path: Path to checkpoint file
            - checkpoint_info: Dict with {name, iteration, loss}
            - config_path: Path to config file
            - logs_dir: Path to TensorBoard logs or None
            - model_config: Model configuration dict
            - loss_history: Dict with loss history

    """
    logger.info("Starting artifacts upload")

    # Unpack session_info
    team_id = session_info['team_id']
    task_id = session_info['task_id']
    project_id = session_info['project_id']
    framework_name = session_info['framework_name']
    task_type = session_info['task_type']
    class_map: ClassMap = session_info['class_map']
    model_meta = sly.ProjectMeta(obj_classes=class_map.obj_classes)
    start_time = session_info['start_time']
    train_size = session_info['train_size']
    initial_samples = session_info.get('initial_samples', 0)

    # Unpack artifacts
    checkpoint_path = artifacts['checkpoint_path']
    checkpoint_info = artifacts['checkpoint_info']
    config_path = artifacts['config_path']
    logs_dir = artifacts.get('logs_dir')
    model_config = artifacts['model_config']

    work_dir = Path(os.path.dirname(checkpoint_path)).parent
    output_dir = work_dir / "upload_artifacts"
    output_dir.mkdir(parents=True, exist_ok=True)

    project_info = api.project.get_info_by_id(project_id)
    project_name = project_info.name if project_info else "unknown"
    model_name = f"Live training - {project_name}"
    model_config['model_name'] = model_name
    remote_dir = f"/experiments/live_training/{project_id}_{project_name}/{task_id}_{framework_name}/"
    logger.info(f"Remote directory: {remote_dir}")

    checkpoints_dir = output_dir / "checkpoints"
    checkpoints_dir.mkdir(exist_ok=True)
    checkpoint_dest = checkpoints_dir / checkpoint_info["name"]
    shutil.copy2(checkpoint_path, checkpoint_dest)

    state_json_src = checkpoint_path.replace('.pth', '_state.json')
    if os.path.exists(state_json_src):
        state_json_dest = str(checkpoint_dest).replace('.pth', '_state.json')
        shutil.copy2(state_json_src, state_json_dest)

    if config_path and os.path.exists(config_path):
        config_dest = output_dir / os.path.basename(config_path)
        shutil.copy2(config_path, config_dest)
        model_files = {"config": config_dest.name}
    else:
        model_files = {}

    sly_json.dump_json_file(
        model_meta.to_json(),
        str(output_dir / "model_meta.json")
    )

    hyperparams = {}
    if config_path and os.path.exists(config_path):
        try:
            hyperparams = LiveTrainingGenerator.parse_hyperparameters(config_path)
        except Exception as e:
            logger.warning(f"Failed to parse hyperparameters: {e}")

    with open(output_dir / "hyperparameters.yaml", 'w') as f:
        yaml.dump(hyperparams, f, default_flow_style=False)

    with open(output_dir / "open_app.lnk", 'w') as f:
        f.write(f"/apps/sessions/{task_id}")

    if logs_dir and os.path.exists(logs_dir):
        logs_dest = output_dir / "logs"
        if logs_dest.exists():
            shutil.rmtree(logs_dest)
        shutil.copytree(logs_dir, logs_dest)
        logger.info(f"Logs copied from {logs_dir}")
        has_logs = True
    else:
        logger.warning("No logs provided")
        has_logs = False

    logger.info("Uploading to Team Files")
    api.file.upload_directory_fast(
        team_id=team_id,
        local_dir=str(output_dir),
        remote_dir=remote_dir,
        change_name_if_conflict=False,
        replace_if_conflict=True
    )

    experiment_info = {
        "experiment_name": f"Live Training {task_type.capitalize()} - Task {task_id}",
        "framework_name": framework_name,
        "model_name": model_name,
        "base_checkpoint": None,
        "base_checkpoint_link": None,
        "task_type": task_type,
        "project_id": project_id,
        "project_version": project_info.version if project_info else None,
        "task_id": task_id,
        "model_files": model_files,
        "checkpoints": [f"checkpoints/{checkpoint_info['name']}"],
        "best_checkpoint": checkpoint_info['name'],
        "export": {},
        "model_meta": "model_meta.json",
        "hyperparameters": "hyperparameters.yaml",
        "hyperparameters_id": None,
        "artifacts_dir": remote_dir,
        "datetime": start_time,
        "experiment_report_id": None,
        "evaluation_report_link": None,
        "evaluation_metrics": {},
        "primary_metric": None,
        "logs": {"type": "tensorboard", "link": f"{remote_dir}logs/"} if has_logs else None,
        "device": get_device_name(),
        "training_duration": calculate_duration(start_time),
        "train_collection_id": None,
        "val_collection_id": None,
        "project_preview": project_info.image_preview_url if project_info else None,
        "train_size": train_size,
        "initial_samples": initial_samples,
        "val_size": 0,
    }

    checkpoints_info = [checkpoint_info]
    loss_history = artifacts.get('loss_history', {})

    session_info = {
        "session_id": task_id,
        "session_name": experiment_info["experiment_name"],
        "project_id": project_id,
        "start_time": experiment_info["datetime"],
        "duration": experiment_info["training_duration"],
        "artifacts_dir": remote_dir,
        "logs_dir": f"{remote_dir}logs/" if has_logs else None,
        "checkpoints": checkpoints_info,
        "loss_history": loss_history,
        "hyperparameters": hyperparams,
        "status": "completed",
        "device": experiment_info["device"],
        "dataset_size": train_size,
        "initial_samples": 0,
        "samples_added": 0,
        "final_size": train_size,
        "train_size": train_size,
        "initial_samples": initial_samples,
        "val_size": 0,
    }

    report_dir = work_dir / "live_training_report"
    report_dir.mkdir(exist_ok=True)

    generator = LiveTrainingGenerator(
        api=api,
        session_info=session_info,
        model_config=model_config,
        model_meta=model_meta,
        output_dir=str(report_dir),
        team_id=team_id,
        task_type=task_type,
    )

    generator.generate()

    # Upload report with overwrite enabled
    report_remote_dir = os.path.join(remote_dir, "visualization")
    file_info = generator.upload(
        remote_dir=report_remote_dir,
        team_id=team_id,
        replace_if_conflict=True,
        change_name_if_conflict=False
    )
    generator._report_file_info = file_info

    report_id = file_info if isinstance(file_info, int) else getattr(file_info, 'id', file_info)
    report_url = f"{api.server_address}/nn/experiments/{report_id}"

    logger.info(f"Report URL: {report_url}")

    experiment_info["has_report"] = True
    experiment_info["experiment_report_id"] = int(report_id)

    # Set output experiment - this REPLACES the entire output
    response = api.task.set_output_experiment(task_id, experiment_info)

    return report_url


def calculate_duration(start_time: str) -> str:
    """Calculate training duration in 'Xh Ym' format."""
    try:
        start_dt = datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S")
        duration_sec = (datetime.now() - start_dt).total_seconds()
        hours = int(duration_sec // 3600)
        minutes = int((duration_sec % 3600) // 60)
        return f"{hours}h {minutes}m"
    except:
        return "N/A"


def get_device_name() -> str:
    """Get GPU device name or 'cpu'."""
    import torch # pylint: disable=import-error
    if not os.path.exists("/dev/nvidia0"):
        return "cpu"

    try:
        if torch.cuda.is_available():
            device_id = int(os.getenv("CUDA_VISIBLE_DEVICES", "0").split(",")[0])
            return torch.cuda.get_device_name(device_id)
    except Exception as e:
        logger.warning(f"Failed to get GPU name: {e}")

    return "cuda"



# coding: utf-8
from copy import deepcopy
import json

from supervisely import logger
from supervisely.annotation.obj_class_collection import ObjClassCollection
from supervisely.annotation.tag_meta_collection import TagMetaCollection
from supervisely.imaging.image import read as sly_image_read
from supervisely.io.json import load_json_file
from supervisely.nn.legacy.hosted.constants import MODEL, SETTINGS, INPUT_SIZE, HEIGHT, WIDTH
from supervisely.nn.legacy.hosted.legacy.inference_config import (
    maybe_convert_from_v1_inference_task_config,
    maybe_convert_from_deploy_task_config,
)
from supervisely.nn.legacy.config import update_recursively
from supervisely.project.project_meta import ProjectMeta
from supervisely.task.paths import TaskPaths
from supervisely.task.progress import Progress
from supervisely.worker_api.interfaces import SingleImageInferenceInterface


GPU_DEVICE = 'gpu_device'


class SingleImageInferenceBase(SingleImageInferenceInterface):
    """Legacy base for single-image inference: loads config, model, and implements inference() and get_out_meta()."""

    def __init__(self, task_model_config=None, _load_model_weights=True):
        """
        :param task_model_config: Override config or None to load from task.
        :type task_model_config: dict
        :param _load_model_weights: If True, load model weights.
        :type _load_model_weights: bool
        """
        logger.info('Starting base single image inference applier init.')
        task_model_config = self._load_task_model_config() if task_model_config is None else deepcopy(task_model_config)
        self._config = update_recursively(self.get_default_config(), task_model_config)
        # Only validate after merging task config with the defaults.
        self._validate_model_config(self._config)

        self._load_train_config()
        if _load_model_weights:
            self._construct_and_fill_model()
        logger.info('Base single image inference applier init done.')

    def _construct_and_fill_model(self):
        progress_dummy = Progress('Building model:', 1)
        progress_dummy.iter_done_report()

    def _validate_model_config(self, config):
        pass

    def inference(self, image, ann):
        raise NotImplementedError()

    def inference_image_file(self, image_file, ann):
        image = sly_image_read(image_file)
        return self.inference(image, ann)

    @staticmethod
    def get_default_config():
        return {}

    @property
    def class_title_to_idx_key(self):
        return 'class_title_to_idx'

    @property
    def train_classes_key(self):
        return 'classes'

    @property
    def model_out_meta(self):
        return self._model_out_meta

    def get_out_meta(self):
        return self._model_out_meta

    def _model_out_tags(self):
        return TagMetaCollection()  # Empty by default

    def _load_raw_model_config_json(self):
        try:
            with open(TaskPaths.MODEL_CONFIG_PATH) as fin:
                self.train_config = json.load(fin)
        except FileNotFoundError:
            raise RuntimeError('Unable to run inference, config from training was not found.')

    @staticmethod
    def _load_task_model_config():
        raw_task_config = load_json_file(TaskPaths.TASK_CONFIG_PATH)
        raw_task_config = maybe_convert_from_deploy_task_config(raw_task_config)
        task_config = maybe_convert_from_v1_inference_task_config(raw_task_config)
        return task_config[MODEL]

    def _load_train_config(self):
        self._load_raw_model_config_json()

        self.class_title_to_idx = self.train_config[self.class_title_to_idx_key]
        logger.info('Read model internal class mapping', extra={'class_mapping': self.class_title_to_idx})
        train_classes = ObjClassCollection.from_json(self.train_config[self.train_classes_key])
        logger.info('Read model out classes', extra={'classes': train_classes.to_json()})

        # TODO: Factor out meta constructing from _load_train_config method.
        self._model_out_meta = ProjectMeta(obj_classes=train_classes, tag_metas=self._model_out_tags())
        # Make a separate [index] --> [class] map that excludes the 'special' classes that should not be in the`
        # final output.
        self.out_class_mapping = {idx: train_classes.get(title) for title, idx in self.class_title_to_idx.items() if
                                  train_classes.has_key(title)}

    def _determine_model_input_size(self):
        src_size = self.train_config[SETTINGS][INPUT_SIZE]
        self.input_size = (src_size[HEIGHT], src_size[WIDTH])
        logger.info('Model input size is read (for auto-rescale).', extra={INPUT_SIZE: {
            WIDTH: self.input_size[1], HEIGHT: self.input_size[0]
        }})



from typing import Dict, Optional, Union

from supervisely.convert.base_converter import AvailablePointcloudEpisodesConverters
from supervisely.convert.pointcloud.bag.bag_converter import BagConverter
from supervisely.convert.pointcloud_episodes.pointcloud_episodes_converter import PointcloudEpisodeConverter
from supervisely.pointcloud_annotation.pointcloud_episode_annotation import PointcloudEpisodeAnnotation
from supervisely.project.project_settings import LabelingInterface


class BagEpisodesConverter(BagConverter, PointcloudEpisodeConverter):
    """Converter for ROS bag files containing point cloud data to use for point cloud episodes."""

    class Item(BagConverter.Item):
        """Episode item: point cloud frame from bag with frame number for episode ordering."""

        def __init__(
            self,
            item_path,
            frame_number: int,
            ann_data: str = None,
            related_images: list = None,
            custom_data: dict = None,
        ):
            """:param frame_number: Frame index in episode."""
            super().__init__(item_path, ann_data, related_images, custom_data)
            self._frame_number = frame_number

        def create_empty_annotation(self) -> PointcloudEpisodeAnnotation:
            return PointcloudEpisodeAnnotation()

    def __init__(
            self,
            input_data: str,
            labeling_interface: Optional[Union[LabelingInterface, str]],
            upload_as_links: bool,
            remote_files_map: Optional[Dict[str, str]] = None,
    ):
        """See :class:`~supervisely.convert.base_converter.BaseConverter` for params."""
        super().__init__(input_data, labeling_interface, upload_as_links, remote_files_map)

        self._type = "point_cloud_episode"
        self._is_pcd_episode = True

    def __str__(self) -> str:
        return AvailablePointcloudEpisodesConverters.BAG



from typing import List, Union

from supervisely.nn.benchmark.semantic_segmentation.base_vis_metric import (
    SemanticSegmVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    MarkdownWidget,
    TableWidget,
)


class Speedtest(SemanticSegmVisMetric):
    """Inference speed and FPS tables/charts for semantic segmentation reports."""


    def is_empty(self) -> bool:
        return self.eval_result.speedtest_info is None

    def multiple_batche_sizes(self) -> bool:
        return len(self.eval_result.speedtest_info["speedtest"]) > 1

    @property
    def latency(self) -> List[Union[int, str]]:
        if self.eval_result.speedtest_info is None:
            return ["N/A"]
        latency = []
        for test in self.eval_result.speedtest_info["speedtest"]:
            latency.append(round(test["benchmark"]["total"], 2))
        return latency

    @property
    def fps(self) -> List[Union[int, str]]:
        if self.eval_result.speedtest_info is None:
            return ["N/A"]
        fps = []
        for test in self.eval_result.speedtest_info["speedtest"]:
            fps.append(round(1000 / test["benchmark"]["total"], 2))
        return fps

    @property
    def intro_md(self) -> MarkdownWidget:
        device = self.eval_result.speedtest_info["model_info"]["device"]
        hardware = self.eval_result.speedtest_info["model_info"]["hardware"]
        runtime = self.eval_result.speedtest_info["model_info"]["runtime"]
        num_it = self.eval_result.speedtest_info["speedtest"][0]["num_iterations"]

        return MarkdownWidget(
            name="speedtest_intro",
            title="Inference Speed",
            text=self.vis_texts.markdown_speedtest_intro.format(device, hardware, runtime, num_it),
        )

    @property
    def intro_table(self) -> TableWidget:
        res = {}

        columns = [" ", "Inference time", "FPS"]
        temp_res = {}
        max_fps = 0
        for test in self.eval_result.speedtest_info["speedtest"]:
            batch_size = test["batch_size"]

            ms = round(test["benchmark"]["total"], 2)
            fps = round(1000 / test["benchmark"]["total"] * batch_size)
            row = [batch_size, ms, fps]
            temp_res[batch_size] = row
            max_fps = max(max_fps, fps)

        res["content"] = []
        # sort by batch size
        temp_res = dict(sorted(temp_res.items()))
        for row in temp_res.values():
            dct = {
                "row": row,
                "id": row[0],
                "items": row,
            }
            res["content"].append(dct)

        columns_options = [
            {"disableSort": True},  # "customCell": True
            {"subtitle": "ms", "tooltip": "Milliseconds for batch images", "postfix": "ms"},
            {
                "subtitle": "imgs/sec",
                "tooltip": "Frames (images) per second",
                "postfix": "fps",
                "maxValue": max_fps,
            },
        ]

        res["columns"] = columns
        res["columnsOptions"] = columns_options

        table = TableWidget(
            name="speedtest_intro_table",
            data=res,
            show_header_controls=False,
            fix_columns=1,
        )
        # table.main_column = "Batch size"
        table.fixed_columns = 1
        table.show_header_controls = False
        return table

    @property
    def batch_size_md(self) -> MarkdownWidget:
        return MarkdownWidget(
            name="batch_size",
            title="Batch Size",
            text=self.vis_texts.markdown_batch_inference,
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget(name="speed_charts", figure=self.get_figure())

    def get_figure(self):  #  -> Optional[go.Figure]
        import plotly.graph_objects as go  # pylint: disable=import-error
        from plotly.subplots import make_subplots  # pylint: disable=import-error

        fig = make_subplots(cols=2)

        ms_color = "#e377c2"
        fps_color = "#17becf"

        temp_res = {}
        for test in self.eval_result.speedtest_info["speedtest"]:
            batch_size = test["batch_size"]

            std = test["benchmark_std"]["total"]
            ms = test["benchmark"]["total"]
            fps = round(1000 / test["benchmark"]["total"] * batch_size)

            ms_line = temp_res.setdefault("ms", {})
            fps_line = temp_res.setdefault("fps", {})
            ms_std_line = temp_res.setdefault("ms_std", {})

            ms_line[batch_size] = ms
            fps_line[batch_size] = fps
            ms_std_line[batch_size] = round(std, 2)

        fig.add_trace(
            go.Scatter(
                x=list(temp_res["ms"].keys()),
                y=list(temp_res["ms"].values()),
                name="Inference time (ms)",
                line=dict(color=ms_color),
                customdata=list(temp_res["ms_std"].values()),
                error_y=dict(
                    type="data",
                    array=list(temp_res["ms_std"].values()),
                    visible=True,
                    color="rgba(227, 119, 194, 0.7)",
                ),
                hovertemplate="Batch Size: %{x}<br>Time: %{y:.2f} ms<br> Standard deviation: %{customdata:.2f} ms<extra></extra>",
            ),
            col=1,
            row=1,
        )
        fig.add_trace(
            go.Scatter(
                x=list(temp_res["fps"].keys()),
                y=list(temp_res["fps"].values()),
                name="FPS",
                line=dict(color=fps_color),
                hovertemplate="Batch Size: %{x}<br>FPS: %{y:.2f}<extra></extra>",
            ),
            col=2,
            row=1,
        )

        fig.update_xaxes(title_text="Batch size", col=1, dtick=1)
        fig.update_xaxes(title_text="Batch size", col=2, dtick=1)

        fig.update_yaxes(title_text="Time (ms)", col=1)
        fig.update_yaxes(title_text="FPS", col=2)
        fig.update_layout(height=400)

        return fig



from typing import Optional, Union, List

from supervisely import ObjClass, ObjClassCollection
from supervisely.app import StateJson
from supervisely.app.widgets import Widget, ObjectClassView


class ClassesListPreview(Widget):
    """Widget that displays a scrollable preview list of ObjClasses with shape icons/titles."""

    def __init__(
        self,
        classes: Optional[Union[List[ObjClass], ObjClassCollection]] = [],
        max_height: str = "128px",
        empty_text: str = None,
        show_shape_title: bool = True,
        show_shape_icon: bool = True,
        widget_id: Optional[str] = None,
    ):
        """
        :param classes: List of ObjClass instances or ObjClassCollection.
        :type classes: Optional[Union[List[ObjClass], ObjClassCollection]]
        :param max_height: Maximum height of the preview area.
        :type max_height: str
        :param empty_text: Text to show when no classes.
        :type empty_text: str, optional
        :param show_shape_title: Whether to show shape type labels.
        :type show_shape_title: bool
        :param show_shape_icon: Whether to show shape icons.
        :type show_shape_icon: bool
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._classes = [
            ObjectClassView(cls, show_shape_title, show_shape_icon).get_json_data()
            for cls in classes
        ]
        self._max_height = max_height
        self._empty_text = empty_text
        self._show_shape_title = show_shape_title
        self._show_shape_icon = show_shape_icon

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "max_height": self._max_height,
        }

    def get_json_state(self):
        return {"classes": [cls for cls in self._classes]}

    def set(
        self,
        classes: Union[List[ObjClass], ObjClassCollection],
        show_shape_title: bool = True,
        show_shape_icon: bool = True,
    ):
        self._classes = [
            ObjectClassView(cls, show_shape_title, show_shape_icon).get_json_data()
            for cls in classes
        ]
        StateJson()[self.widget_id] = self.get_json_state()
        StateJson().send_changes()

    def get(self):
        return self._classes



"""Functions for processing pointcloud episodes"""

from supervisely._utils import abs_url, is_development


def get_labeling_tool_url(dataset_id, pointcloud_id):
    """
    Get the URL for the labeling tool with the specified dataset ID and poin tcloud ID.

    :param dataset_id: Dataset ID in Supervisely.
    :type dataset_id: int
    :param pointcloud_id: Point cloud ID in Supervisely.
    :type pointcloud_id: int
    :returns: URL for the labeling tool with the specified dataset ID and point cloud ID
    :rtype: str

    :Usage Example:

        .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
                load_dotenv(os.path.expanduser("~/supervisely.env"))

            api = sly.Api.from_env()

            pointcloud_id = 19373403
            pcd_info = api.pointcloud_episodes.get_info_by_id(pointcloud_id)
            url = sly.pointcloud_episodes.get_labeling_tool_url(pcd_info.dataset_id, pcd_info.id)

            print(url)
            # Output:
            # https://app.supervisely.com/app/point-clouds-tracking/?datasetId=55875&pointCloudId=19373403
    """

    res = f"/app/point-clouds-tracking/?datasetId={dataset_id}&pointCloudId={pointcloud_id}"
    if is_development():
        res = abs_url(res)
    return res


def get_labeling_tool_link(url, name="open in labeling tool"):
    """
    Get HTML link to the labeling tool with the specified URL and name.

    :param url: URL of the labeling tool.
    :type url: str
    :param name: Name of the link, default is "open in labeling tool".
    :type name: str
    :returns: HTML link to the labeling tool with the specified URL and name.
    :rtype: str

    :Usage Example:

        .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
                load_dotenv(os.path.expanduser("~/supervisely.env"))

            api = sly.Api.from_env()

            pointcloud_id = 19373403
            pcd_info = api.pointcloud_episodes.get_info_by_id(pointcloud_id)
            url = sly.pointcloud_episodes.get_labeling_tool_url(pcd_info.dataset_id, pcd_info.id)
            name = "my link"

            link = sly.pointcloud_episodes.get_labeling_tool_link(url, name)

            print(link)
            # Output:
            # <a
            #     href="https://app.supervisely.com/app/point-clouds/?datasetId=55875&pointCloudId=19373403"
            #     rel="noopener noreferrer"
            #     target="_blank"
            # >
            #     my link<i class="zmdi zmdi-open-in-new" style="margin-left: 5px"></i>
            # </a>
    """

    return f'<a href="{url}" rel="noopener noreferrer" target="_blank">{name}<i class="zmdi zmdi-open-in-new" style="margin-left: 5px"></i></a>'



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from typing import List
import supervisely as sly
from supervisely.app import StateJson
from supervisely.app.widgets import Widget


class SelectAppSession(Widget):
    """Widget to select an app session by tags within a team (returns selected session ID)."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        team_id: int,
        tags: List[str],
        show_label: bool = False,
        size: Literal["large", "small", "mini"] = None,
        widget_id: str = None,
        operation: str = "or",
    ):
        """:param team_id: Team ID to list sessions from.
        :type team_id: int
        :param tags: List of tag names to filter sessions.
        :type tags: List[str]
        :param show_label: If True, show label.
        :type show_label: bool
        :param size: Size: "large", "small", or "mini".
        :type size: Literal["large", "small", "mini"], optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        :param operation: Tag filter: "or" or "and".
        :type operation: str

        :raises ValueError: If tags is not a non-empty list.
        """
        self._session_id = None
        self._team_id = team_id
        self._tags = tags
        self._show_label = show_label
        self._size = size
        self._operation = operation

        if not isinstance(tags, list):
            raise ValueError("Parameter tags must be a list of strings")
        
        if len(tags) < 1:
            raise ValueError("Parameter tags must be a list of strings, but got empty list")

        super().__init__(widget_id=widget_id, file_path=__file__)

    def value_changed(self, func):
        route_path = self.get_route_path(SelectAppSession.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_selected_id()
            func(res)

        return _click

    def get_json_data(self):
        data = {}
        data["teamId"] = self._team_id
        data["ssOptions"] = {
            "sessionTags": self._tags,
            "showLabel": self._show_label,
            "sessionTagsOperation": self._operation,
        }
        if self._size is not None:
            data["ssOptions"]["size"] = self._size
        return data

    def get_json_state(self):
        state = {}
        state["sessionId"] = self._session_id
        return state

    def set_session_id(self, session_id: int):
        self._session_id = session_id
        StateJson()[self.widget_id]["sessionId"] = self._session_id
        StateJson().send_changes()

    def get_selected_id(self):
        return StateJson()[self.widget_id]["sessionId"]



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class DialogWindowBase(Exception):
    """Base exception that can be displayed to the user as an in-app dialog window."""

    def __init__(self, title, description, status):
        """
        :param title: Dialog title.
        :type title: str
        :param description: Dialog description/message.
        :type description: str
        :param status: Status: "info", "success", "warning", or "error".
        :type status: str
        """
        self.title = title
        self.description = description
        self.status = status
        super().__init__(self.get_message())

    def get_message(self):
        return f"{self.title}: {self.description}"

    def __str__(self):
        return self.get_message()


class DialogWindowError(DialogWindowBase):
    """Dialog window exception with `error` status."""

    def __init__(self, title, description):
        """
        :param title: Dialog title.
        :type title: str
        :param description: Error message.
        :type description: str
        """
        super().__init__(title, description, "error")


class DialogWindowWarning(DialogWindowBase):
    """Dialog window exception with `warning` status."""

    def __init__(self, title, description):
        """
        :param title: Dialog title.
        :type title: str
        :param description: Warning message.
        :type description: str
        """
        super().__init__(title, description, "warning")


# for compatibility
class DialogWindowMessage(DialogWindowError):
    """Backwards-compatible alias for informational dialog exceptions."""

    def __init__(self, title, description):
        """
        :param title: Dialog title.
        :type title: str
        :param description: Info message.
        :type description: str
        """
        super().__init__(title, description, "info")  # pylint: disable=too-many-function-args


def show_dialog(
    title, description, status: Literal["info", "success", "warning", "error"] = "info"
):
    """
    Show a dialog window with the given title and message.

    :param title: Dialog title.
    :type title: str
    :param description: Dialog description/message.
    :type description: str
    :param status: Status: "info", "success", "warning", or "error".
    :type status: Literal["info", "success", "warning", "error"]
    """
    from supervisely.app import StateJson, DataJson

    StateJson()["slyAppShowDialog"] = True
    DataJson()["slyAppDialogTitle"] = title
    DataJson()["slyAppDialogMessage"] = description
    DataJson()["slyAppDialogStatus"] = status
    DataJson().send_changes()
    StateJson().send_changes()



import supervisely.io.env as sly_env


class Singleton(type):
    """Metaclass that enforces single instance per class (with per-user variants for DataJson/StateJson in multiuser mode)."""

    _instances = {}
    _nested_instances = {}

    def __call__(cls, *args, **kwargs):
        local = kwargs.pop("__local__", False)
        if local is False:
            if cls not in cls._instances:
                cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)

            if sly_env.is_multiuser_mode_enabled():
                from supervisely.app.content import DataJson, StateJson
                from copy import deepcopy

                # Initialize nested instances dict once
                nested_instances = cls._nested_instances.setdefault(cls, {})
                
                user_id = sly_env.user_from_multiuser_app()
                if user_id is not None and cls in (StateJson, DataJson):
                    if user_id not in nested_instances:
                        # Create new instance and copy data
                        instance = super(Singleton, cls).__call__(*args, **kwargs)
                        instance.update(deepcopy(dict(cls._instances[cls])))
                        nested_instances[user_id] = instance
                    
                    return nested_instances[user_id]
            return cls._instances[cls]
        else:
            return super(Singleton, cls).__call__(*args, **kwargs)



# coding: utf-8
"""Create and manage users in your Supervisely team."""

# docs
from __future__ import annotations

from collections import namedtuple
from typing import TYPE_CHECKING, Callable, Dict, List, NamedTuple, Optional, Union

from supervisely.api.module_api import ApiField, ModuleApiBase, _get_single_item

if TYPE_CHECKING:
    from pandas.core.frame import DataFrame

from tqdm import tqdm


class UserInfo(NamedTuple):
    """NamedTuple describing a user returned by the API."""

    id: int
    login: str
    role: str
    role_id: int
    name: str
    email: str
    logins: int
    disabled: bool
    last_login: str
    created_at: str
    updated_at: str


class UserApi(ModuleApiBase):
    """API for working with users."""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                api = sly.Api.from_env()
                users = api.user.get_list()
        """
        super().__init__(api)

    Membership = namedtuple("Membership", ["id", "name", "role_id", "role"])

    @staticmethod
    def info_sequence():
        """
        Sequence of fields that are returned by the API to represent UserInfo.

        :Usage Example:

            .. code-block:: python

                UserInfo(
                    id=8,
                    login="alex",
                    role=None,
                    role_id=None,
                    name=None,
                    email=None,
                    logins=20,
                    disabled=False,
                    last_login="2021-03-24T15:06:26.804Z",
                    created_at="2020-04-17T10:24:09.077Z",
                    updated_at="2021-03-24T15:13:01.148Z",
                )
        """
        return [
            ApiField.ID,
            ApiField.LOGIN,
            ApiField.ROLE,
            ApiField.ROLE_ID,
            ApiField.NAME,
            ApiField.EMAIL,
            ApiField.LOGINS,
            ApiField.DISABLED,
            ApiField.LAST_LOGIN,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Name of the tuple that represents UserInfo.
        """
        return "UserInfo"

    def _convert_json_info(self, info: dict, skip_missing=True):
        res = super(UserApi, self)._convert_json_info(info, skip_missing=skip_missing)
        return UserInfo(**res._asdict())

    def get_info_by_id(self, id: int) -> UserInfo:
        """
        Get User information by User ID.

        :param id: User ID in Supervisely.
        :type id: int
        :returns: Information about User.
        :rtype: :class:`~supervisely.api.user_api.UserInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                user_info = api.user.get_info_by_id(8)
                print(user_info)
                # Output: [
                #     8,
                #     "alex",
                #     null,
                #     null,
                #     null,
                #     null,
                #     20,
                #     false,
                #     "2021-03-24T15:06:26.804Z",
                #     "2020-04-17T10:24:09.077Z",
                #     "2021-03-24T15:13:01.148Z"
                # ]
        """
        return self._get_info_by_id(id, "users.info")

    def get_info_by_login(self, login: str) -> UserInfo:
        """
        Get User information by User login.

        :param login: User login in Supervisely.
        :type login: str
        :returns: Information about User.
        :rtype: :class:`~supervisely.api.user_api.UserInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                user_info = api.user.get_info_by_login('alex')
                print(user_info)
                # Output: [
                #     8,
                #     "alex",
                #     null,
                #     null,
                #     null,
                #     null,
                #     20,
                #     false,
                #     "2021-03-24T15:06:26.804Z",
                #     "2020-04-17T10:24:09.077Z",
                #     "2021-03-24T15:13:01.148Z"
                # ]
        """
        filters = [{"field": ApiField.LOGIN, "operator": "=", "value": login}]
        items = self.get_list(filters)
        return _get_single_item(items)

    def get_member_info_by_login(self, team_id: int, login: str) -> UserInfo:
        """
        Get information about team member by Team ID and User login by login.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param login: User login in Supervisely.
        :type login: str
        :returns: Information about User.
        :rtype: :class:`~supervisely.api.user_api.UserInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                member_info = api.user.get_member_info_by_login(64, 'alex')
                print(member_info)
                # Output: [
                #     8,
                #     "alex",
                #     "manager",
                #     3,
                #     null,
                #     null,
                #     20,
                #     false,
                #     "2021-03-24T15:06:26.804Z",
                #     "2020-04-17T10:24:09.077Z",
                #     "2021-03-24T15:13:01.148Z"
                # ]
        """
        filters = [{"field": ApiField.LOGIN, "operator": "=", "value": login}]
        team_members = self.get_list_all_pages(
            "members.list",
            {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters},
            convert_json_info_cb=self._api.user._convert_json_info,
        )
        return _get_single_item(team_members)

    def get_member_info_by_id(self, team_id: int, user_id: int) -> UserInfo:
        """
        Get information about team member by Team ID and User ID.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param user_id: User ID in Supervisely.
        :type user_id: int
        :returns: Information about User.
        :rtype: :class:`~supervisely.api.user_api.UserInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                member_info = api.user.get_member_info_by_id(64, 8)
                print(member_info)
                # Output: [
                #     8,
                #     "alex",
                #     "manager",
                #     3,
                #     null,
                #     null,
                #     20,
                #     false,
                #     "2021-03-24T15:06:26.804Z",
                #     "2020-04-17T10:24:09.077Z",
                #     "2021-03-24T15:13:01.148Z"
                # ]
        """
        filters = [{"field": ApiField.ID, "operator": "=", "value": user_id}]
        team_members = self.get_list_all_pages(
            "members.list",
            {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters},
            convert_json_info_cb=self._api.user._convert_json_info,
        )
        return _get_single_item(team_members)

    def get_list(self, filters: List[Dict[str, str]] = None) -> List[UserInfo]:
        """
        Get list of information about Users.

        :param filters: List of params to sort output Users.
        :type filters: List[dict], optional
        :returns: List of information about Users.
        :rtype: :class:`List[UserInfo]`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                # Get list of Users with id = 8
                user_info = api.user.get_list(filters=[{'field': 'id', 'operator': '=', 'value': '8'}])
                print(user_info)
                # Output: [
                #     8,
                #     "alex",
                #     "manager",
                #     3,
                #     null,
                #     null,
                #     20,
                #     false,
                #     "2021-03-24T15:06:26.804Z",
                #     "2020-04-17T10:24:09.077Z",
                #     "2021-03-24T15:13:01.148Z"
                # ]
        """
        return self.get_list_all_pages("users.list", {ApiField.FILTER: filters or []})

    def create(
        self,
        login: str,
        password: str,
        is_restricted: Optional[bool] = False,
        name: Optional[str] = "",
        email: Optional[str] = "",
    ) -> UserInfo:
        """
        Creates new User with given login and password.

        :param login: New User login.
        :type login: str
        :param password: New User password.
        :type password: str
        :param is_restricted: If True, new User will have no access to Explore section, won't be able to create or switch Teams, and no personal Team will be created for this User during signup.
        :type is_restricted: bool, optional
        :param name: New User name.
        :type name: str, optional
        :param email: New User email.
        :type email: str, optional
        :returns: Information about new User.
        :rtype: :class:`~supervisely.api.user_api.UserInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                new_user_info = api.user.create('John', 'qwerty', is_restricted=True, name='John Wick', email='excomunicado@gmail.com')
                print(new_user_info)
                # Output: [
                #     274,
                #     "John",
                #     null,
                #     null,
                #     "John Wick",
                #     "excomunicado@gmail.com",
                #     0,
                #     false,
                #     null,
                #     "2021-03-24T16:20:03.110Z",
                #     "2021-03-24T16:20:03.110Z"
                # ]
        """
        response = self._api.post(
            "users.add",
            {
                ApiField.LOGIN: login,
                ApiField.PASSWORD: password,
                ApiField.IS_RESTRICTED: is_restricted,
                ApiField.NAME: name,
                ApiField.EMAIL: email,
            },
        )
        return self.get_info_by_id(response.json()[ApiField.USER_ID])

    def _set_disabled(self, id, disable):
        """
        Check status of the user with given id
        :param id: int
        :param disable: bool
        """
        self._api.post("users.disable", {ApiField.ID: id, ApiField.DISABLE: disable})

    def disable(self, id: int) -> None:
        """
        Disables User with the given ID.

        :param id: User ID in Supervisely.
        :type id: int
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                user_id = 8
                api.user.disable(user_id)
        """
        self._set_disabled(id, True)

    def enable(self, id: int) -> None:
        """
        Enables User with the given ID.

        :param id: User ID in Supervisely.
        :type id: int
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                user_id = 8
                api.user.enable(user_id)
        """
        self._set_disabled(id, False)

    def get_token(self, login):
        raise NotImplementedError()

    def get_teams(self, id: int) -> List[UserInfo]:
        """
        Get list with information about User Teams.

        :param id: User ID in Supervisely.
        :type id: int
        :returns: List of teams where the user is a member (team id, name, and role).
        :rtype: List[:class:`~supervisely.api.user_api.UserInfo`]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                teams = api.user.get_teams(8)
                print(teams)
                # Output: [
                #     [
                #         9,
                #         "alex",
                #         1,
                #         "admin"
                #     ],
                #     [
                #         64,
                #         "test",
                #         3,
                #         "manager"
                #     ]
                # ]
        """
        response = self._api.post("users.info", {ApiField.ID: id})
        teams_json = response.json()[ApiField.TEAMS]
        teams = []
        for team in teams_json:
            member = self.Membership(
                id=team[ApiField.ID],
                name=team[ApiField.NAME],
                role_id=team[ApiField.ROLE_ID],
                role=team[ApiField.ROLE],
            )
            teams.append(member)
        return teams

    def add_to_team(self, user_id: int, team_id: int, role_id: int) -> None:
        """
        Invites User to Team with the given role.

        :param user_id: User ID in Supervisely.
        :type user_id: int
        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param role_id: Role ID in Supervisely.
        :type role_id: int
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                user_id = 8
                team_id = 76
                role_id = 5
                api.user.add_to_team(user_id, team_id, role_id)
        """
        user = self.get_info_by_id(user_id)
        self._api.post(
            "members.add",
            {
                ApiField.LOGIN: user.login,
                ApiField.TEAM_ID: team_id,
                ApiField.ROLE_ID: role_id,
            },
        )

    def remove_from_team(self, user_id: int, team_id: int) -> None:
        """
        Removes User from Team.

        :param user_id: User ID in Supervisely.
        :type user_id: int
        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                user_id = 8
                team_id = 76
                api.user.remove_from_team(user_id, team_id)
        """
        self._api.post("members.remove", {ApiField.ID: user_id, ApiField.TEAM_ID: team_id})

    def update(
        self, id: int, password: Optional[str] = None, name: Optional[str] = None
    ) -> UserInfo:
        """
        Updates User info.

        :param id: User ID in Supervisely.
        :type id: int
        :param password: User password.
        :type password: str
        :param name: User name.
        :type name: str
        :returns: New information about User.
        :rtype: :class:`~supervisely.api.user_api.UserInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                user_info = api.user.update(8, name='Aleksey')
                print(user_info)
                # Output: [
                #     8,
                #     "alex",
                #     null,
                #     null,
                #     "Aleksey",
                #     null,
                #     21,
                #     false,
                #     "2021-03-25T08:06:03.498Z",
                #     "2020-04-17T10:24:09.077Z",
                #     "2021-03-25T08:37:17.257Z"
                # ]
        """
        data = {}
        if password is not None:
            data[ApiField.PASSWORD] = password
        if name is not None:
            data[ApiField.NAME] = name
        if len(data) == 0:
            return
        data[ApiField.ID] = id

        self._api.post("users.editInfo", data)
        return self.get_info_by_id(id)

    def change_team_role(self, user_id: int, team_id: int, role_id: int) -> None:
        """
        Changes User role in the given Team.

        :param user_id: User ID in Supervisely.
        :type user_id: int
        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param role_id: Role ID in Supervisely.
        :type role_id: int
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                user_id = 8
                team_id = 64
                new_role_id = 2
                api.user.change_team_role(user_id, team_id, new_role_id)
        """
        self._api.post(
            "members.editInfo",
            {
                ApiField.ID: user_id,
                ApiField.TEAM_ID: team_id,
                ApiField.ROLE_ID: role_id,
            },
        )

    def get_team_members(self, team_id: int) -> List[UserInfo]:
        """
        Get list of information about Team Users.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :returns: List of information about Team Users
        :rtype: :class:`List[UserInfo]`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                team_id = 9
                team_members = api.user.get_team_members(team_id)
        """
        team_members = self.get_list_all_pages(
            "members.list",
            {ApiField.TEAM_ID: team_id, ApiField.FILTER: []},
            convert_json_info_cb=self._api.user._convert_json_info,
        )
        return team_members

    def get_team_role(self, user_id: int, team_id: int) -> UserInfo:
        """
        Get Team role for given User and Team IDs.

        :param user_id: User ID in Supervisely.
        :type user_id: int
        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :returns: Membership information for the specified team (team id, name, and role).
        :rtype: :class:`~supervisely.api.user_api.UserInfo`, optional

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                user_id = 8
                team_id = 9
                team_role = api.user.get_team_role(user_id, team_id)
                print(team_role)
                # Output: TeamMembership(id=9, name="alex", role_id=1, role="admin")
        """
        user_teams = self.get_teams(user_id)
        for member in user_teams:
            if member.id == team_id:
                return member
        return None

    def get_member_activity(
        self, team_id: int, user_id: int, progress_cb: Optional[Union[tqdm, Callable]] = None
    ) -> DataFrame:
        """
        Get User activity data.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param user_id: User ID in Supervisely.
        :type user_id: int
        :param progress_cb: Function to check progress.
        :type progress_cb: tqdm or callable, optional
        :returns: Activity data as `pd.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_
        :rtype: :class:`pd.DataFrame`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                activity = api.user.get_member_activity(64, 8)
                print(activity)
                # Output:
                #    userId               action                      date  ... jobId   tag tagId
                # 0        8        login_to_team  2021-03-13T08:57:26.832Z  ...  None  None  None
                # 1        8  annotation_duration  2021-03-02T13:16:23.833Z  ...  None  None  None
                # 2        8        login_to_team  2021-03-02T13:15:58.775Z  ...  None  None  None
                # 3        8        login_to_team  2021-02-06T09:47:22.999Z  ...  None  None  None
                # ................................................................................
                # 38       8     create_workspace  2021-01-04T12:25:37.916Z  ...  None  None  None
                # 39       8        login_to_team  2021-01-04T12:24:58.257Z  ...  None  None  None
                # 40       8        login_to_team  2021-01-04T12:23:43.056Z  ...  None  None  None
                # 41       8        login_to_team  2021-01-04T11:53:56.447Z  ...  None  None  None
                # [42 rows x 18 columns]
        """
        import pandas as pd

        activity = self._api.team.get_activity(
            team_id, filter_user_id=user_id, progress_cb=progress_cb
        )
        df = pd.DataFrame(activity)
        return df

    def add_to_team_by_login(self, user_login: str, team_id: int, role_id: int) -> Dict[str, int]:
        """
        Invite User to Team with given role by login.

        :param user_login: User login in Supervisely.
        :type user_login: str
        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param role_id: Role ID in Supervisely.
        :type role_id: int
        :returns: Information about new User in Team
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                team_id = 13
                role_id = 2
                new_user_data = api.user.add_to_team_by_login('alex', team_id, role_id)
                print(new_user_data)
                # Output: {"userId": 8}
        """
        response = self._api.post(
            "members.add",
            {
                ApiField.LOGIN: user_login,
                ApiField.TEAM_ID: team_id,
                ApiField.ROLE_ID: role_id,
            },
        )
        return response.json()

    def get_ssh_keys(self):
        """ """
        response = self._api.post("users.ssh-keys", {})
        return response.json()

    def get_my_info(self):
        response = self._api.post("users.me", {})
        return self._convert_json_info(response.json())



# coding: utf-8
"""
Single prediction returned by deployed model.
"""

from __future__ import annotations

import atexit
import os
import tempfile
from os import PathLike
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
import requests

from supervisely._utils import get_valid_kwargs, logger, rand_str
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_meta import TagValueType
from supervisely.api.api import Api
from supervisely.convert.image.sly.sly_image_helper import get_meta_from_annotation
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging.image import read as read_image
from supervisely.imaging.image import read_bytes as read_image_bytes
from supervisely.imaging.image import write as write_image
from supervisely.io.fs import (
    clean_dir,
    dir_empty,
    ensure_base_path,
    get_file_ext,
    mkdir,
)
from supervisely.project.project_meta import ProjectMeta
from supervisely.video.video import VideoFrameReader


class Prediction:
    """
    A single prediction result.

    The prediction is primarily represented by ``annotation_json`` (Supervisely annotation format).
    If you provide ``model_meta`` (a :class:`~supervisely.project.project_meta.ProjectMeta` or its JSON),
    you can access the parsed :class:`~supervisely.annotation.annotation.Annotation` via
    :attr:`~supervisely.nn.model.prediction.Prediction.annotation`.
    """
    _temp_dir = os.path.join(tempfile.gettempdir(), "prediction_files")
    __cleanup_registered = False

    def __init__(
        self,
        annotation_json: Dict,
        source: Union[str, int] = None,
        model_meta: Optional[Union[ProjectMeta, Dict]] = None,
        name: Optional[str] = None,
        path: Optional[str] = None,
        url: Optional[str] = None,
        project_id: Optional[int] = None,
        dataset_id: Optional[int] = None,
        image_id: Optional[int] = None,
        video_id: Optional[int] = None,
        frame_index: Optional[int] = None,
        api: Optional["Api"] = None,
        **kwargs,
    ):
        """
        :param annotation_json: Prediction in Supervisely annotation JSON format.
        :type annotation_json: dict or :class:`~supervisely.annotation.annotation.Annotation`
        :param source: Optional source descriptor (path, id, etc.), used as a hint for loading/visualization.
        :type source: str or int, optional
        :param model_meta: Output meta of the model, required to construct an :class:`~supervisely.annotation.annotation.Annotation`.
        :type model_meta: :class:`~supervisely.project.project_meta.ProjectMeta` or dict, optional
        :param name: Optional name for this prediction (used in visualization filenames).
        :type name: str, optional
        :param path: Local path to the source image/video.
        :type path: str, optional
        :param url: URL to the source image/video.
        :type url: str, optional
        :param project_id: Project id associated with the source.
        :type project_id: int, optional
        :param dataset_id: Dataset id associated with the source.
        :type dataset_id: int, optional
        :param image_id: Image id associated with the source.
        :type image_id: int, optional
        :param video_id: Video id associated with the source (for frame predictions).
        :type video_id: int, optional
        :param frame_index: Frame index for video predictions.
        :type frame_index: int, optional
        :param api: API client used to download image/frame by id.
        :type api: :class:`~supervisely.api.api.Api`, optional
        """
        self.source = source
        if isinstance(annotation_json, Annotation):
            annotation_json = annotation_json.to_json()

        self.annotation_json = annotation_json
        self.model_meta = model_meta
        if isinstance(self.model_meta, dict):
            self.model_meta = ProjectMeta.from_json(self.model_meta)

        self.name = name
        self.path = path
        self.url = url
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.image_id = image_id
        self.video_id = video_id
        self.frame_index = frame_index
        self.extra_data = {}
        if kwargs:
            self.extra_data.update(kwargs)
        self.api = api

        self._annotation = None
        self._boxes = None
        self._masks = None
        self._classes = None
        self._scores = None
        self._track_ids = None

        if self.path is None and isinstance(self.source, (str, PathLike)):
            self.path = str(self.source)

    def _init_geometries(self):
        """Initialize the geometries of the prediction."""
        def _get_confidence(label: Label):
            for tag_name in ["confidence", "conf", "score"]:
                conf_tag: Tag = label.tags.get(tag_name, None)
                if conf_tag is not None and conf_tag.meta.value_type != str(
                    TagValueType.ANY_NUMBER
                ):
                    conf_tag = None
                if conf_tag is not None:
                    return conf_tag.value
            return 1

        self._boxes = []
        self._masks = []
        self._classes = []
        self._scores = []
        for label in self.annotation.labels:
            self._classes.append(label.obj_class.name)
            self._scores.append(_get_confidence(label))
            if isinstance(label.geometry, Rectangle):
                rect = label.geometry
            else:
                rect = label.geometry.to_bbox()
            if isinstance(label.geometry, Bitmap):
                self._masks.append(label.geometry.get_mask(self.annotation.img_size))

            self._boxes.append(
                np.array(
                    [
                        rect.top,
                        rect.left,
                        rect.bottom,
                        rect.right,
                    ]
                )
            )
        self._boxes = np.array(self._boxes)
        self._masks = np.array(self._masks)

        custom_data = self.annotation.custom_data
        if custom_data and isinstance(custom_data, list) and len(custom_data) == len(self.annotation.labels):
            self._track_ids = np.array(custom_data)

    @property
    def boxes(self):
        """Get the bounding boxes of the prediction."""
        if self._boxes is None:
            self._init_geometries()
        return self._boxes

    @property
    def masks(self):
        """Get the masks of the prediction."""
        if self._masks is None:
            self._init_geometries()
        return self._masks

    @property
    def classes(self):
        """Get the classes of the prediction."""
        if self._classes is None:
            self._init_geometries()
        return self._classes

    @property
    def scores(self):
        """Get the scores of the prediction."""
        if self._scores is None:
            self._init_geometries()
        return self._scores

    @property
    def annotation(self) -> Annotation:
        """Get the annotation of the prediction."""
        if self._annotation is None and self.annotation_json is not None:
            if self.model_meta is None:
                raise ValueError("Model meta is not provided. Cannot create annotation.")
            model_meta = get_meta_from_annotation(self.annotation_json, self.model_meta)
            self._annotation = Annotation.from_json(self.annotation_json, model_meta)
        return self._annotation

    @annotation.setter
    def annotation(self, annotation: Union[Annotation, Dict]):
        """Set the annotation of the prediction."""
        if isinstance(annotation, Annotation):
            self._annotation = annotation
            self.annotation_json = annotation.to_json()
        elif isinstance(annotation, dict):
            self._annotation = None
            self.annotation_json = annotation
        else:
            raise ValueError("Annotation must be either a dict or an Annotation object.")

    @property
    def class_idxs(self) -> np.ndarray:
        """Get the class indexes of the prediction."""
        if self.model_meta is None:
            raise ValueError("Model meta is not provided. Cannot create class indexes.")
        cls_name_to_idx = {
            obj_class.name: i for i, obj_class in enumerate(self.model_meta.obj_classes)
        }
        return np.array([cls_name_to_idx[class_name] for class_name in self.classes])
    @property
    def track_ids(self):
        """Get track IDs for each detection. Returns None for detections without tracking.

        :returns: Track IDs for each detection.
        :rtype: np.ndarray
        """
        if self._track_ids is None:
            self._init_geometries()
        return self._track_ids

    @classmethod
    def from_json(cls, json_data: Dict, **kwargs) -> "Prediction":
        """
        Create :class:`~supervisely.nn.model.prediction.Prediction` from a JSON dict.

        This helper accepts both styles of keys coming from different backends:
        ``annotation_json`` or ``annotation``.

        :param json_data: Source JSON.
        :type json_data: dict
        :param kwargs: Additional fields to override/extend JSON (e.g. ``api``, ``model_meta``).
        :type kwargs: dict
        :returns: Prediction instance.
        :rtype: :class:`~supervisely.nn.model.prediction.Prediction`
        :raises ValueError: If annotation payload is missing.
        """

        kwargs = {**json_data, **kwargs}
        if "annotation_json" in kwargs:
            annotation_json = kwargs.pop("annotation_json")
        elif "annotation" in kwargs:
            annotation_json = kwargs.pop("annotation")
        else:
            raise ValueError("Annotation JSON is required.")
        kwargs = get_valid_kwargs(
            kwargs,
            Prediction.__init__,
            exclude=["self", "annotation_json"],
        )
        return cls(annotation_json, **kwargs)

    def to_json(self):
        """
        Serialize prediction to a JSON-compatible dict.

        :returns: JSON dict.
        :rtype: dict
        """

        return {
            "source": self.source,
            "annotation": self.annotation_json,
            "name": self.name,
            "path": self.path,
            "url": self.url,
            "project_id": self.project_id,
            "dataset_id": self.dataset_id,
            "image_id": self.image_id,
            "video_id": self.video_id,
            "frame_index": self.frame_index,
            **self.extra_data,
        }

    def _clear_temp_files(self):
        if not dir_empty(self._temp_dir):
            clean_dir(self._temp_dir)

    def load_image(self) -> np.ndarray:
        """
        Load the source image (or video frame) into a NumPy array.

        The loader uses the first available source in the following order:
        ``path`` → ``url`` → ``image_id`` (requires ``api`` or environment config) →
        ``video_id`` + ``frame_index`` → video ``path``/``url`` + ``frame_index``.

        :returns: Image (or frame) as ``H×W×C`` NumPy array.
        :rtype: numpy.ndarray
        """

        api = self.api
        if self.frame_index is None:
            if self.path is not None:
                return read_image(self.path)
            if self.url is not None:
                ext = get_file_ext(self.url)
                if ext == "":
                    ext = ".jpg"
                r = requests.get(self.url, stream=True, timeout=60)
                r.raise_for_status()
                return read_image_bytes(r.content)
            if self.image_id is not None:
                try:
                    if api is None:
                        # TODO: raise more clarifying error in case of failing of api init
                        # what a user should do to fix it?
                        api = Api()
                    return api.image.download_np(self.image_id)
                except Exception as e:
                    raise RuntimeError("Failed to load image by ID") from e
            raise ValueError("Cannot load image. No path, URL, image ID, or frame_index provided.")
        if self.video_id is not None:
            if self.frame_index is None:
                raise ValueError("Frame index is not provided for video.")
            try:
                if api is None:
                    api = Api()
                return api.video.frame.download_np(self.video_id, self.frame_index)
            except Exception as e:
                raise RuntimeError("Failed to load frame by video ID") from e
        if self.path is not None:
            return next(VideoFrameReader(self.path, frame_indexes=[self.frame_index]))
        if self.url is not None:
            video_name = Path(self.url).name
            video_path = Path(self._temp_dir) / video_name
            mkdir(self._temp_dir)
            if not video_path.exists():
                with requests.get(self.url, stream=True, timeout=10 * 60) as r:
                    r.raise_for_status()
                    with open(video_path, "wb") as f:
                        for chunk in r.iter_content(chunk_size=8192):
                            f.write(chunk)
            return next(VideoFrameReader(video_path, frame_indexes=[self.frame_index]))
        raise ValueError("Cannot load frame. No path, URL or video ID provided.")

    def visualize(
        self,
        save_path: Optional[str] = None,
        save_dir: Optional[str] = None,
        color: Optional[List[int]] = None,
        thickness: Optional[int] = None,
        opacity: Optional[float] = 0.5,
        draw_tags: Optional[bool] = False,
        fill_rectangles: Optional[bool] = True,
    ) -> np.ndarray:
        """
        Render prediction on top of the source image/frame.

        If ``save_path`` (file or directory) or ``save_dir`` is provided, the visualization is saved.
        Otherwise the rendered image is returned.

        :param save_path: Output file path or directory.
        :type save_path: str, optional
        :param save_dir: Output directory (alias for directory ``save_path``).
        :type save_dir: str, optional
        :param color: Optional RGB color for drawing.
        :type color: List[int], optional
        :param thickness: Optional line thickness.
        :type thickness: int, optional
        :param opacity: Bitmap opacity.
        :type opacity: float, optional
        :param draw_tags: If True, draws tags on objects.
        :type draw_tags: bool, optional
        :param fill_rectangles: If True, fills rectangles.
        :type fill_rectangles: bool, optional
        :returns: Rendered image.
        :rtype: numpy.ndarray
        """

        if save_dir is not None and save_path is not None:
            raise ValueError("Only one of save_path or save_dir can be provided.")

        mkdir(self._temp_dir)
        if not Prediction.__cleanup_registered:
            atexit.register(self._clear_temp_files)
            Prediction.__cleanup_registered = True

        img = self.load_image()
        self.annotation.draw_pretty(
            bitmap=img,
            color=color,
            thickness=thickness,
            opacity=opacity,
            draw_tags=draw_tags,
            output_path=None,
            fill_rectangles=fill_rectangles,
        )
        if save_dir is not None:
            save_path = save_dir
        if save_path is None:
            return img
        if Path(save_path).suffix == "":
            # is a directory
            if self.name is not None:
                name = self.name
                if self.frame_index is not None:
                    name = f"{name}_{self.frame_index}"
            elif self.image_id is not None:
                name = str(self.image_id)
            elif self.video_id is not None:
                name = str(self.video_id)
            elif self.path is not None:
                name = Path(self.path).name
                if self.frame_index is not None:
                    name = f"{name}_{self.frame_index}"
            else:
                name = f"{rand_str(6)}"
            name = f"vis_{name}.png"
            save_path = os.path.join(save_path, name)
        ensure_base_path(save_path)
        write_image(save_path, img)
        logger.info("Visualization for prediction saved to %s", save_path)
        return img



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from supervisely.api.api import Api


class Video(Widget):
    """Video player widget that can be controlled from Python and emits playback/frame events."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        PLAY_CLICKED = "play_clicked_cb"
        PAUSE_CLICKED = "pause_clicked_cb"
        FRAME_CHANGE_START = "frame_change_started_cb"
        FRAME_CHANGE_END = "frame_change_finished_cb"

    def __init__(
        self,
        video_id: int = None,
        # intervals: List[List[int]] = [],
        widget_id: str = None,
    ):
        """:param video_id: Video ID from Supervisely to play.
        :type video_id: int, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._api = Api()
        self._video_id = video_id
        self._video_info = None
        if self._video_id is not None:
            self._video_info = self._api.video.get_info_by_id(self._video_id, raise_error=True)

        self._intervals = []
        self._loading: bool = False

        self._play_clicked_handled = False
        self._pause_clicked_handled = False
        self._frame_change_started_handled = False
        self._frame_change_finished_handled = False

        #############################
        # video settings
        self._sound_volume: int = 1
        self._playback_rate: int = 1
        self._skip_frames_size: int = 10
        self._intervals_navigation: bool = False
        self._responsive_height: bool = True
        self._enable_zoom: bool = False
        #############################
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "videoId": self._video_id,
            "intervals": self._intervals,
            "loading": self._loading,
            "options": {
                "soundVolume": self._sound_volume,
                "playbackRate": self._playback_rate,
                "skipFramesSize": self._skip_frames_size,
                "intervalsNavigation": self._intervals_navigation,
                "responsiveHeight": self._responsive_height,
                "enableZoom": self._enable_zoom,
            },
        }

    def get_json_state(self):
        return {"currentFrame": 0, "startFrame": 0}

    @property
    def video_id(self):
        return self._video_id

    def set_video(self, id: int):
        self._video_id = id
        self._video_info = self._api.video.get_info_by_id(self._video_id, raise_error=True)
        DataJson()[self.widget_id]["videoId"] = self._video_id
        DataJson()[self.widget_id]["frames_count"] = self._video_info.frames_count
        DataJson().send_changes()
        StateJson()[self.widget_id]["currentFrame"] = 0
        StateJson().send_changes()

    @property
    def loading(self):
        return self._loading

    @loading.setter
    def loading(self, value: bool):
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading
        DataJson().send_changes()

    def set_current_frame(self, value):
        if self._video_info is None:
            raise ValueError("VideoID is not defined yet, use 'set_video' method")
        if value >= self.get_frames_count():
            value = self.get_frames_count() - 1
        StateJson()[self.widget_id]["currentFrame"] = value
        StateJson().send_changes()

    def get_current_frame(self):
        if self._video_info is None:
            raise ValueError("VideoID is not defined yet, use 'set_video' method")
        return max(0, int(StateJson()[self.widget_id]["currentFrame"]))

    def play_clicked(self, func):
        route_path = self.get_route_path(Video.Routes.PLAY_CLICKED)
        server = self._sly_app.get_server()
        self._play_clicked_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_current_frame()
            func(res)

        return _click

    def pause_clicked(self, func):
        route_path = self.get_route_path(Video.Routes.PAUSE_CLICKED)
        server = self._sly_app.get_server()
        self._pause_clicked_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_current_frame()
            func(res)

        return _click

    def frame_change_started(self, func):
        route_path = self.get_route_path(Video.Routes.FRAME_CHANGE_START)
        server = self._sly_app.get_server()
        self._frame_change_started_handled = True

        @server.post(route_path)
        def _click():
            res = max(0, StateJson()[self.widget_id]["startFrame"])
            func(res)

        return _click

    def frame_change_finished(self, func):
        route_path = self.get_route_path(Video.Routes.FRAME_CHANGE_END)
        server = self._sly_app.get_server()
        self._frame_change_finished_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_current_frame()
            func(res)

        return _click

    def get_frames_count(self):
        if self._video_info is None:
            raise ValueError("VideoID is not defined yet, use 'set_video' method")
        return self._video_info.frames_count



from supervisely import logger
from supervisely.nn.legacy.dataset import ensure_samples_nonempty
from supervisely.nn.legacy.hosted.class_indexing import CONTINUE_TRAINING, TRANSFER_LEARNING
from supervisely.nn.legacy.hosted.pytorch.constants import CUSTOM_MODEL_CONFIG, HEAD_LAYER
from supervisely.nn.legacy.hosted.trainer import (
    SuperviselyModelTrainer,
    BATCH_SIZE,
    DATASET_TAGS,
    EPOCHS,
    LOSS,
    LR,
    TRAIN,
    VAL,
    WEIGHTS_INIT_TYPE,
    INPUT_SIZE,
    HEIGHT,
    WIDTH,
)
from supervisely.nn.legacy.training.eval_planner import EvalPlanner, VAL_EVERY
from supervisely.nn.legacy.pytorch.dataset import PytorchSegmentationSlyDataset
from supervisely.nn.legacy.pytorch.weights import WeightsRW
from supervisely.task.paths import TaskPaths
from supervisely.task.progress import Progress, epoch_float, report_metrics_training, report_metrics_validation

# pylint: disable=import-error
import torch
from torch.autograd import Variable
from torch.utils.data import DataLoader
# pylint: enable=import-error

def _check_all_pixels_have_segmentation_class(targets):
    if targets.min().item() < 0:
        raise ValueError('Training data has an item that is not fully covered by the segmentation labeling '
                         '(some image pixels are not assigned any class). This is an error. If you need to '
                         'ignore some parts of the input images, use a special dummy class and a custom loss '
                         'function that would ignore pixels of that class.')


class PytorchSegmentationTrainer(SuperviselyModelTrainer):
    """Legacy PyTorch semantic segmentation trainer: trains model, saves checkpoints, reports metrics."""

    @staticmethod
    def get_default_config():
        return {
            INPUT_SIZE: {
                WIDTH: 256,
                HEIGHT: 256
            },
            DATASET_TAGS: {
                TRAIN: 'train',
                VAL: 'val',
            },
            BATCH_SIZE: {
                TRAIN: 1,
                VAL: 1,
            },
            EPOCHS: 2,
            VAL_EVERY: 0.5,
            LR: 0.001,
            WEIGHTS_INIT_TYPE: TRANSFER_LEARNING,  # CONTINUE_TRAINING,
            CUSTOM_MODEL_CONFIG: {}    # Model-specific settings go in this section.
        }

    def __init__(self, model_factory_fn, optimization_loss_fn, training_metrics_dict=None, default_model_config=None):
        """
        :param model_factory_fn: Model factory.
        :type model_factory_fn: Callable
        :param optimization_loss_fn: Loss function.
        :type optimization_loss_fn: Callable
        :param training_metrics_dict: Optional metrics.
        :type training_metrics_dict: dict
        :param default_model_config: Optional model config.
        :type default_model_config: dict
        """
        default_config = PytorchSegmentationTrainer.get_default_config()
        default_config[CUSTOM_MODEL_CONFIG].update(default_model_config if default_model_config is not None else {})

        self._model_factory_fn = model_factory_fn
        self._optimization_loss_fn = optimization_loss_fn
        self._training_metrics_dict = training_metrics_dict.copy() if training_metrics_dict is not None else {}
        if LOSS in self._training_metrics_dict:
            raise ValueError('aaaaaaa')
        self._metrics_with_loss = self._training_metrics_dict.copy()
        self._metrics_with_loss[LOSS] = self._optimization_loss_fn

        super().__init__(default_config=default_config)

    def get_start_class_id(self):
        """Set the integer segmentation class indices to start from 0.

        Some segmentation neural net implementations treat class id 0 in a special way (e.g. as a background class) and
        need the indexing to start from 1.
        """
        return 0

    def _determine_model_classes(self):
        """Look at input dataset segmentation classes and assign integer class ids."""

        # This also automatically reuses an existing class id mapping in continue_training mode (continuing training
        # from a previous snapshot with a dataset fully compatible in terms of classes).
        super()._determine_model_classes_segmentation(bkg_input_idx=-1)

    def _construct_and_fill_model(self):
        # Progress reporting to show a progress bar in the UI.
        model_build_progress = Progress('Building model:', 1)

        # Check the class name --> index mapping to infer the number of model output dimensions.
        num_classes = max(self.class_title_to_idx.values()) + 1

        # Initialize the model.
        model = self._model_factory_fn(
            num_classes=num_classes,
            input_size=self._input_size,
            custom_model_config=self.config.get(CUSTOM_MODEL_CONFIG, {}))
        logger.info('Model has been instantiated.')

        # Load model weights appropriate for the given training mode.
        weights_rw = WeightsRW(TaskPaths.MODEL_DIR)
        weights_init_type = self.config[WEIGHTS_INIT_TYPE]
        if weights_init_type == TRANSFER_LEARNING:
            # For transfer learning, do not attempt to load the weights for the model head. The existing snapshot may
            # have been trained on a different dataset, even on a different set of classes, and is in general not
            # compatible with the current model even in terms of dimensions. The head of the model will be initialized
            # randomly.
            self._model = weights_rw.load_for_transfer_learning(model, ignore_matching_layers=[HEAD_LAYER],
                                                                logger=logger)
        elif weights_init_type == CONTINUE_TRAINING:
            # Continuing training from an older snapshot requires full compatibility between the two models, including
            # class index mapping. Hence the snapshot weights must exactly match the structure of our model instance.
            self._model = weights_rw.load_strictly(model)

        # Model weights have been loaded, move them over to the GPU.
        self._model.cuda()

        # Advance the progress bar and log a progress message.
        logger.info('Weights have been loaded.', extra={WEIGHTS_INIT_TYPE: weights_init_type})
        model_build_progress.iter_done_report()

    def _construct_loss(self):
        pass

    def _construct_data_loaders(self):
        # Initialize the IO logic to feed the model during training.

        # Dimensionality of all images in an input batch must be the same.
        # We fix the input size for the whole dataset. Every image will be scaled to this size before feeding the model.
        src_size = self.config[INPUT_SIZE]
        input_size = (src_size[HEIGHT], src_size[WIDTH])

        # We need separate data loaders for the training and validation folds.
        self._data_loaders = {}

        # The train dataset should be re-shuffled every epoch, but the validation dataset samples order is fixed.
        for dataset_name, need_shuffle, drop_last in [(TRAIN, True, True), (VAL, False, False)]:
            # For more informative logging, grab the tag marking the respective dataset images.
            dataset_tag = self.config[DATASET_TAGS][dataset_name]

            # Get a list of samples for the dataset in question, make sure it is not empty.
            samples = self._samples_by_data_purpose[dataset_name]
            ensure_samples_nonempty(samples, dataset_tag, self.project.meta)

            # Instantiate the dataset object to handle sample indexing and image resizing.
            dataset = PytorchSegmentationSlyDataset(
                project_meta=self.project.meta,
                samples=samples,
                out_size=input_size,
                class_mapping=self.class_title_to_idx,
                bkg_color=-1
            )
            # Report progress.
            logger.info('Prepared dataset.', extra={
                'dataset_purpose': dataset_name, 'tag': dataset_tag, 'samples': len(samples)
            })
            # Initialize a PyTorch data loader. For the training dataset, set the loader to ignore the last incomplete
            # batch to avoid noisy gradients and batchnorm updates.
            self._data_loaders[dataset_name] = DataLoader(
                dataset=dataset,
                batch_size=self.config[BATCH_SIZE][dataset_name],
                shuffle=need_shuffle,
                drop_last=drop_last
            )

        # Report progress
        logger.info('DataLoaders have been constructed.')

        # Compute the number of iterations per epoch for training and validation.
        self._train_iters = len(self._data_loaders[TRAIN])
        self._val_iters = len(self._data_loaders[VAL])
        self._epochs = self.config[EPOCHS]

        # Initialize a helper to determine when to pause training and perform validation and snapshotting.
        self._eval_planner = EvalPlanner(epochs=self._epochs, val_every=self.config[VAL_EVERY])

    def _dump_model_weights(self, out_dir):
        # Framework-specific logic to snapshot the model weights.
        WeightsRW(out_dir).save(self._model)

    def _validation(self):
        # Compute validation metrics.

        # Switch the model to evaluation model to stop batchnorm runnning average updates.
        self._model.eval()
        # Initialize the totals counters.
        validated_samples = 0
        total_val_metrics = {name: 0.0 for name in self._metrics_with_loss}
        total_loss = 0.0

        # Iterate over validation dataset batches.
        for val_it, (inputs, targets) in enumerate(self._data_loaders[VAL]):
            _check_all_pixels_have_segmentation_class(targets)

            # Move the data to the GPU and run inference.
            with torch.no_grad():
                inputs_cuda, targets_cuda = Variable(inputs).cuda(), Variable(targets).cuda()

            outputs_cuda = self._model(inputs_cuda)

            # The last betch may be smaller than the rest if the dataset does not have a whole number of full batches,
            # so read the batch size from the input.
            batch_size = inputs_cuda.size(0)

            # Compute the metrics and grab the values from GPU.
            batch_metrics = {name: metric_fn(outputs_cuda, targets_cuda).item()
                             for name, metric_fn in self._metrics_with_loss.items()}
            for name, metric_value in batch_metrics.items():
                total_val_metrics[name] += metric_value * batch_size

            # Add up the totals.
            validated_samples += batch_size

            # Report progress.
            logger.info("Validation in progress",
                        extra={'epoch': self.epoch_flt, 'val_iter': val_it, 'val_iters': self._val_iters})

        # Compute the average loss from the accumulated totals.
        avg_metrics_values = {name: total_value / validated_samples for name, total_value in total_val_metrics.items()}

        # Report progress and metric values to be plotted in the training chart and return.
        report_metrics_validation(self.epoch_flt, avg_metrics_values)
        logger.info("Validation has been finished", extra={'epoch': self.epoch_flt})
        return avg_metrics_values

    def train(self):
        # Initialize the progesss bar in the UI.
        training_progress = Progress('Model training: ', self._epochs * self._train_iters)

        # Initialize the optimizer.
        optimizer = torch.optim.Adam(self._model.parameters(), lr=self.config[LR])
        # Running best loss value to determine which snapshot is the best so far.
        best_val_loss = float('inf')

        for epoch in range(self._epochs):
            logger.info("Starting new epoch", extra={'epoch': self.epoch_flt})
            for train_it, (inputs_cpu, targets_cpu) in enumerate(self._data_loaders[TRAIN]):
                _check_all_pixels_have_segmentation_class(targets_cpu)

                # Switch the model into training mode to enable gradient backpropagation and batch norm running average
                # updates.
                self._model.train()

                # Copy input batch to the GPU, run inference and compute optimization loss.
                inputs_cuda, targets_cuda = Variable(inputs_cpu).cuda(), Variable(targets_cpu).cuda()
                outputs_cuda = self._model(inputs_cuda)
                loss = self._optimization_loss_fn(outputs_cuda, targets_cuda)

                # Make a gradient descent step.
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                metric_values = {name: metric_fn(outputs_cuda, targets_cuda).item()
                                 for name, metric_fn in self._training_metrics_dict.items()}
                metric_values[LOSS] = loss.item()

                # Advance UI progess bar.
                training_progress.iter_done_report()
                # Compute fractional epoch value for more precise metrics reporting.
                self.epoch_flt = epoch_float(epoch, train_it + 1, self._train_iters)
                # Report metrics to be plotted in the training chart.
                report_metrics_training(self.epoch_flt, metric_values)

                # If needed, do validation and snapshotting.
                if self._eval_planner.need_validation(self.epoch_flt):
                    # Compute metrics on the validation dataset.
                    metrics_values_val = self._validation()

                    # Report progress.
                    self._eval_planner.validation_performed()

                    # Check whether the new weights are the best so far on the validation dataset.
                    val_loss = metrics_values_val[LOSS]
                    model_is_best = val_loss < best_val_loss
                    if model_is_best:
                        best_val_loss = val_loss

                    # Save a snapshot with the current weights. Mark whether the snapshot is the best so far in terms of
                    # validation loss.
                    self._save_model_snapshot(model_is_best, opt_data={
                        'epoch': self.epoch_flt,
                        'val_metrics': metrics_values_val,
                    })

            # Report progress
            logger.info("Epoch has finished", extra={'epoch': self.epoch_flt})



# coding: utf-8
from __future__ import annotations

from typing import Dict, List, NamedTuple, Optional

from supervisely._utils import batched
from supervisely.api.module_api import ApiField, ModuleApi, RemoveableBulkModuleApi
from supervisely.video_annotation.key_id_map import KeyIdMap


class ObjectApi(RemoveableBulkModuleApi):
    """
    Object for :class:`~supervisely.video_annotation.video_annotation.VideoAnnotation`.
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple ObjectInfo information about Object.

        :Example:

         .. code-block:: python

            ObjectInfo(id=152118,
                       description='',
                       created_at='2021-03-23T13:25:34.705Z',
                       updated_at='2021-03-23T13:25:34.705Z',
                       dataset_id=466642, class_id=2856942,
                       entity_id=198703211,
                       tags=[{'objectId': 152118, 'tagId': 29098694, 'entityId': None, 'id': 40632, 'value': 'grey'}],
                       meta={},
                       created_by_id=16154)
        """
        return [
            ApiField.ID,
            ApiField.DESCRIPTION,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.DATASET_ID,
            ApiField.CLASS_ID,
            ApiField.ENTITY_ID,
            ApiField.TAGS,
            ApiField.META,
            ApiField.CREATED_BY_ID,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Get string name of NamedTuple for class.

        :returns: NamedTuple name.
        :rtype: str

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                tuple_name = api.video.object.info_tuple_name()
                print(tuple_name) # ObjectInfo
        """

        return "ObjectInfo"

    def get_info_by_id(self, id: int) -> NamedTuple:
        """
        Get Object information by ID.

        :param id: Object ID in Supervisely.
        :type id: int
        :returns: Information about Object.
        :rtype: NamedTuple
        """
        return self._get_info_by_id(id, "annotation-objects.info")

    def get_list(
        self, dataset_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[NamedTuple]:
        """
        Get list of information about all video Objects for a given dataset ID.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param filters: List of parameters to sort output Objects.
        :type filters: List[dict], optional
        :returns: Information about Objects.
        :rtype: List[NamedTuple]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                dataset_id = 466642
                object_infos = api.video.object.get_list(dataset_id)
                print(object_infos)
                # Output: [
                #     [
                #         152118,
                #         "",
                #         "2021-03-23T13:25:34.705Z",
                #         "2021-03-23T13:25:34.705Z",
                #         466642,
                #         2856942,
                #         198703211,
                #         [
                #             {
                #                 "objectId": 152118,
                #                 "tagId": 29098694,
                #                 "entityId": null,
                #                 "id": 40632,
                #                 "value": "grey"
                #             }
                #         ],
                #         {},
                #         16154
                #     ],
                #     [
                #         152119,
                #         "",
                #         "2021-03-23T13:25:34.705Z",
                #         "2021-03-23T13:25:34.705Z",
                #         466642,
                #         2856942,
                #         198703211,
                #         [
                #             {
                #                 "objectId": 152119,
                #                 "tagId": 29098694,
                #                 "entityId": null,
                #                 "id": 40633,
                #                 "value": "wine"
                #             }
                #         ],
                #         {},
                #         16154
                #     ],
                #     [
                #         152120,
                #         "",
                #         "2021-03-23T13:25:34.705Z",
                #         "2021-03-23T13:25:34.705Z",
                #         466642,
                #         2856942,
                #         198703211,
                #         [
                #             {
                #                 "objectId": 152120,
                #                 "tagId": 29098694,
                #                 "entityId": null,
                #                 "id": 40634,
                #                 "value": "beige"
                #             }
                #         ],
                #         {},
                #         16154
                #     ],
                #     [
                #         152121,
                #         "",
                #         "2021-03-23T13:25:34.705Z",
                #         "2021-03-23T13:25:34.705Z",
                #         466642,
                #         2856941,
                #         198703212,
                #         [
                #             {
                #                 "objectId": 152121,
                #                 "tagId": 29098696,
                #                 "entityId": null,
                #                 "id": 40635,
                #                 "value": "juvenile"
                #             }
                #         ],
                #         {},
                #         16154
                #     ],
                #     [
                #         152122,
                #         "",
                #         "2021-03-23T13:25:34.705Z",
                #         "2021-03-23T13:25:34.705Z",
                #         466642,
                #         2856943,
                #         198703211,
                #         [],
                #         {},
                #         16154
                #     ]
                # ]
        """
        return self.get_list_all_pages(
            "annotation-objects.list",
            {ApiField.DATASET_ID: dataset_id, ApiField.FILTER: filters or []},
        )

    def _append_bulk(
        self,
        tag_api,
        entity_id,
        project_id,
        dataset_id,
        objects,
        key_id_map: KeyIdMap = None,
        is_pointcloud=False,
        is_video_multi_view: bool = False,
    ):
        """"""
        if len(objects) == 0:
            return []

        objcls_name_id_map = self._api.object_class.get_name_to_id_map(project_id)

        items = []
        for obj in objects:
            new_obj = {ApiField.CLASS_ID: objcls_name_id_map[obj.obj_class.name]}

            if not is_video_multi_view and not is_pointcloud:
                new_obj[ApiField.ENTITY_ID] = entity_id
            items.append(new_obj)

        response = self._api.post(
            "annotation-objects.bulk.add",
            {ApiField.DATASET_ID: dataset_id, ApiField.ANNOTATION_OBJECTS: items},
        )
        ids = [obj[ApiField.ID] for obj in response.json()]
        KeyIdMap.add_objects_to(key_id_map, [obj.key() for obj in objects], ids)

        # add tags to objects
        tag_api.append_to_objects(entity_id, project_id, objects, key_id_map, is_video_multi_view)

        return ids

    def remove_batch(self, ids, progress_cb=None, batch_size=50):
        """
        Remove objects in batches from the Supervisely server.
        All entity IDs must belong to the same item (for example image, volume).

        :param ids: IDs of objects in Supervisely.
        :type ids: List[int]
        :param progress_cb: Function for control remove progress.
        :type progress_cb: Callable

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                object_ids = [19369645, 19369646, 19369647]
                api.volume.object.remove_batch(object_ids)
        """
        for ids_batch in batched(ids, batch_size=batch_size):
            self._api.post(
                "annotation-objects.bulk.remove",
                ids_batch,
            )
            if progress_cb is not None:
                progress_cb(len(ids_batch))
